import yaml
from copy import deepcopy
import matplotlib.pyplot as plt
import networkx as nx
from typing import List, Union, Dict, Optional, Any
import rospy
from eagerx.utils.utils import (
    get_opposite_msg_cls,
    get_module_type_string,
    get_cls_from_string,
    get_attribute_from_module,
    substitute_args,
    msg_type_error,
    supported_types,
)
from eagerx.utils.network_utils import (
    reset_graph,
    episode_graph,
    plot_graph,
    color_nodes,
    color_edges,
    is_stale,
)
from eagerx.core.entities import Node, BaseConverter, SpaceConverter
from eagerx.core.view import GraphView
from eagerx.core.specs import (
    BaseNodeSpec,
    ObjectSpec,
    ConverterSpec,
    EntitySpec,
    NodeSpec,
)

yaml.Dumper.ignore_aliases = lambda *args: True  # todo: check if needed.


def merge(a, b, path=None):
    "merges b into a"
    # If it is a spec, convert to params
    if path is None:
        path = []
    for key in b:
        if isinstance(b[key], EntitySpec):
            b[key] = b[key].params
        if isinstance(b[key], GraphView):
            b[key] = b[key].to_dict()
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge(a[key], b[key], path + [str(key)])
            elif a[key] == b[key]:
                pass  # same leaf value
            else:
                a[key] = b[key]
        else:
            a[key] = b[key]
    return a


class Graph:
    def __init__(self, state: Dict):
        self._state = state

    def __str__(self):
        return yaml.dump(self._state)

    @classmethod
    def create(
        cls,
        nodes: Optional[List[BaseNodeSpec]] = None,
        objects: Optional[List[ObjectSpec]] = None,
    ):
        if nodes is None:
            nodes = []
        if objects is None:
            objects = []
        if isinstance(nodes, BaseNodeSpec):
            nodes = [nodes]
        if isinstance(objects, ObjectSpec):
            objects = [objects]

        # Add action & observation node to list
        import eagerx.core.nodes  # noqa  Required so that actions & observations node are registered.

        actions = Node.make("Actions")
        observations = Node.make("Observations")
        nodes += [actions, observations]

        # Create a state
        state = dict(nodes=dict(), connects=list(), backup=dict(), gui_state=dict())
        graph = cls(state)
        graph.add(nodes)
        graph.add(objects)
        return graph

    def add(
        self,
        entities: Union[Union[BaseNodeSpec, ObjectSpec], List[Union[BaseNodeSpec, ObjectSpec]]],
    ):
        """Add a node/object to the provided state."""
        if not isinstance(entities, list):
            entities = [entities]

        for entity in entities:
            name = entity.config.name
            assert (
                name not in self._state["nodes"]
            ), f'There is already a node or object registered in this graph with name "{name}".'
            assert not entity.has_graph, f"Spec '{name}' is already added to a graph."

            # Add node to state
            self._state["nodes"][name] = entity.params
            self._state["backup"][name] = entity.params

            # Add graph reference to spec
            entity.set_graph(self)

    def remove(self, names: Union[str, EntitySpec, List[Union[str, EntitySpec]]], remove: bool = False):
        """Removes a node.
        First removes all associated connects from self._state.
        Then, removes node/object from self._state.
        Also removes observation entries if they are disconnected when remove=True.
        Also removes action entries if they are disconnect and the last connection when remove=True.
        """
        if not isinstance(names, list):
            names = [names]
        for name in names:
            if isinstance(name, EntitySpec):
                name = name.params["config"]["name"]
                assert name in self._state["nodes"], f" No entity with name '{name}' in graph."
            for source, target in deepcopy(self._state["connects"]):
                if name in [source[0], target[0]]:
                    if source[0] == "env/actions":
                        action = source[2]
                        source = None
                    else:
                        action = None
                        source = self.get_view(source[0], source[1:])
                    if target[0] == "env/observations":
                        observation = target[2]
                        target = None
                    else:
                        observation = None
                        target = self.get_view(target[0], target[1:])
                    self.disconnect(source, target, action, observation, remove=remove)
            self._state["nodes"].pop(name)

    def add_component(
        self,
        entry: Optional[GraphView] = None,
        action: Optional[str] = None,
        observation: Optional[str] = None,
    ):
        """Adds a component entry to the selection list.
        For actions, it will also add an entry in self._state['nodes'][env/actions]['params'][outputs]
        For observations, it will also add an entry in self._state['nodes'][env/observations]['params'][inputs]
        """
        # assert only action, only observation, only entry
        self._correct_signature(entry, action, observation)
        if entry is not None:  # component parameter
            self._add_component(entry)
        if action:
            self._add_action(action)
        if observation:
            self._add_observation(observation)

    def _add_component(self, entry: GraphView):
        """Adds a component entry to the selection list."""
        name, component, cname = entry()

        # For feedthroughs, add the corresponding output instead
        component = "outputs" if component == "feedthroughs" else component

        # Add cname to selection list if it is not already selected
        params = self._state["nodes"][name]
        assert cname not in params["config"][component], '"%s" already selected in "%s" under %s.' % (
            cname,
            name,
            component,
        )
        params["config"][component].append(cname)

    def _add_action(self, action: str):
        """Adds disconnected action entry to 'env/actions' node in self._state."""
        assert action != "set", 'Cannot define an action with the reserved name "set".'
        params_action = self._state["nodes"]["env/actions"]
        if action not in params_action["outputs"]:  # Action already registered
            params_action["outputs"][action] = dict()
            view = self.get_view("env/actions", ["outputs", action])
            self._add_component(view)

    def _add_observation(self, observation: str):
        """Adds disconnected observation entry to 'env/observations' node in self._state."""
        assert observation != "actions_set", 'Cannot define an observations with the reserved name "actions_set".'
        params_obs = self._state["nodes"]["env/observations"]
        if observation in params_obs["inputs"]:
            assert len(params_obs["inputs"][observation]) == 0, (
                'Observation "%s" already exists' " and is connected." % observation
            )
        else:
            params_obs["inputs"][observation] = dict()
            view = self.get_view("env/observations", ["inputs", observation])
            self._add_component(view)

    def remove_component(
        self,
        entry: Optional[GraphView] = None,
        action: Optional[str] = None,
        observation: Optional[str] = None,
        remove: bool = False,
    ):
        """Removes a component entry from the selection list. It will first disconnect all connections in connect.
        For feedthroughs, it will remove the corresponding output from the selection list.
        For actions, it will also remove the entry in self._state['nodes'][env/actions]['params'][outputs]
        For observations, it will also remove the entry in self._state['nodes'][env/observations]['params'][inputs]
        """
        # assert only action, only observation, only name, component, cname
        self._correct_signature(entry, action, observation)
        if entry:  # component parameter
            self._remove_component(entry, remove=remove)
        if action:
            self._remove_action(action)
        if observation:
            self._remove_observation(observation)

    def _remove_component(self, entry: GraphView, remove: bool = False):
        """Removes a component entry from the selection list. It will first disconnect all connections in connect.
        For feedthroughs, it will remove the corresponding output from the selection list.
        """
        # For feedthroughs, remove the corresponding output instead
        name, component, cname = entry()
        component = "outputs" if component == "feedthroughs" else component
        entry = self.get_view(name, [component, cname])
        self._is_selected(self._state, entry)

        # Disconnect component entry
        self._disconnect_component(entry, remove=remove)

        # Remove cname from selection list
        params = self._state["nodes"][name]
        params["config"][component].remove(cname)

    def _remove_action(self, action: str):
        """Method to remove an action. It will first disconnect all connections in connect."""
        view = self.get_view("env/actions", ["outputs", action])
        self._remove_component(view, remove=False)
        params_action = self._state["nodes"]["env/actions"]
        source = ["env/actions", "outputs", action]
        connect_exists = False
        for c in self._state["connects"]:
            if source == c[0]:
                connect_exists = True
                target = c[1]
                break
        assert (
            not connect_exists
        ), f'Action entry "{action}" cannot be removed, because it is not disconnected. Connection with target {target} still exists.'
        params_action["outputs"].pop(action)

    def _remove_observation(self, observation: str):
        """Method to remove an observation. It will first disconnect all connections in connect."""
        view = self.get_view("env/observations", ["inputs", observation])
        self._remove_component(view, remove=False)
        params_obs = self._state["nodes"]["env/observations"]
        target = ["env/observations", "inputs", observation]
        connect_exists = False
        for c in self._state["connects"]:
            if target == c[1]:
                connect_exists = True
                source = c[0]
                break
        assert not connect_exists, (
            'Observation entry "%s" cannot be removed, because it is not disconnected. Connection with source %s still exists.'
            % (observation, source)
        )
        # assert observation in params_obs['inputs'], 'Observation "%s" cannot be removed,
        # because it does not exist.' % observation
        params_obs["inputs"].pop(observation)

    def connect(
        self,
        source: GraphView = None,
        target: GraphView = None,
        action: str = None,
        observation: str = None,
        converter: Optional[Dict] = None,
        window: Optional[int] = None,
        delay: Optional[float] = None,
        skip: Optional[bool] = None,
    ):
        assert not source or not action, (
            'You cannot specify a source if you wish to connect action "%s",' " as the action will act as the source." % action
        )
        assert not target or not observation, (
            'You cannot specify a target if you wish to connect observation "%s",'
            " as the observation will act as the target." % observation
        )
        assert not (observation and action), "You cannot connect an action directly to an observation."

        if isinstance(converter, ConverterSpec):
            converter = converter.params
        if isinstance(converter, GraphView):
            converter = converter.to_dict()

        if action:  # source = action
            self.add_component(action=action)
            self._connect_action(action, target, converter=converter)
            source = self.get_view("env/actions", ["outputs", action])
        elif observation:  # target = observation
            self.add_component(observation=observation)
            converter = self._connect_observation(source, observation, converter=converter)
            target = self.get_view("env/observations", ["inputs", observation])
        self._connect(source, target, converter, window, delay, skip)

    def _connect(
        self,
        source: GraphView = None,
        target: GraphView = None,
        converter: Optional[Dict] = None,
        window: Optional[int] = None,
        delay: Optional[float] = None,
        skip: Optional[bool] = None,
    ):
        """Method to connect a source to a target. For actions/observations, first a (new) disconnected entry must be created,
        after which an additional call to connect_action/observation is required before calling this method.
        For more info, see self.connect."""
        # Perform checks on source
        self._is_selected(self._state, source)

        # Perform checks on target
        target_name, target_comp, target_cname = target()
        if target_comp == "feedthroughs":
            assert window is None or window > 0, "Feedthroughs must have a window > 0, " "else no action can be fed through."
            assert not skip, (
                "Feedthroughs cannot skip, as they only feedthrough outputs. "
                "When setting skip=True, no msg can be fed through."
            )
            ft_view = self.get_view(target_name, ["outputs", target_cname])
            self._is_selected(self._state, ft_view)
        else:
            self._is_selected(self._state, target)

        # Make sure that target is not already connected.
        for s, t in self._state["connects"]:
            t_name, t_comp, t_cname = t
            flag = not (target_name == t_name and target_comp == t_comp and target_cname == t_cname)
            assert flag, f'Target "{target}" is already connected to source "{s}"'

        # Add properties to target params
        if converter is not None:
            self.set({"converter": converter}, target)
        if window is not None:
            self.set({"window": window}, target)
        if delay is not None:
            self.set({"delay": delay}, target)
        if skip is not None:
            self.set({"skip": skip}, target)

        # Add connection
        Graph.check_msg_type(source, target, self._state)
        connect = [list(source()), list(target())]
        self._state["connects"].append(connect)

    def _connect_action(self, action, target, converter=None):
        """Method to connect a (previously added) action, that *precedes* self._connect(source, target)."""
        params_action = self._state["nodes"]["env/actions"]
        assert action in params_action["outputs"], f'Action "{action}" must be added, before you can connect it.'
        name, component, cname = target()
        if component == "feedthroughs":
            component = "outputs"
            target = self.get_view(name, [component, cname])

        # Infer source properties (converter & msg_type) from target
        sc = target.space_converter
        space_converter = sc.to_dict() if isinstance(sc, GraphView) else sc
        msg_type_C = get_cls_from_string(target.msg_type)
        if converter:  # Overwrite msg_type_B if input converter specified
            msg_type_B = get_opposite_msg_cls(msg_type_C, converter)
        else:
            msg_type_B = msg_type_C

        # Set properties in node params of 'env/actions'
        if len(params_action["outputs"][action]) > 0:  # Action already registered
            space_converter_state = params_action["outputs"][action]["converter"]
            msg_type_B_state = get_opposite_msg_cls(params_action["outputs"][action]["msg_type"], space_converter_state)
            assert msg_type_B == msg_type_B_state, (
                f'Conflicting msg_types for action "{action}" that is already '
                f"used in another connection. Occurs with connection %s" % tuple([name, component, cname])
            )

            if not space_converter == space_converter_state:
                rospy.logwarn(
                    'Conflicting %s for action "%s". '
                    "Not using the space_converter "
                    "of %s[%s][%s]" % ("space_converters", action, name, component, cname)
                )
            msg_type_A = get_opposite_msg_cls(msg_type_B_state, space_converter_state)
        else:
            assert target.space_converter is not None, (
                f'"{cname}" does not have a space_converter defined for ' f'{component} in the spec of object "{name}".'
            )
            # Verify that converter is not modifying the msg_type (i.e. it is a processor).
            assert msg_type_B == msg_type_C, (
                "Cannot have a converter that maps to a different msg_type as the "
                "converted msg_type will not be compatible with the "
                "space_converter specified in the spec."
            )
            msg_type_A = get_opposite_msg_cls(msg_type_B, space_converter)
            msg_type = get_module_type_string(msg_type_A)
            mapping = dict(
                msg_type=msg_type,
                rate="$(config rate)",
                converter=space_converter,
                space_converter=None,
            )
            self._set(params_action["outputs"], {action: mapping})

    def _connect_observation(self, source, observation, converter):
        """Method to connect a (previously added & disconnected) observation, that *precedes* self._connect(source, target)."""
        params_obs = self._state["nodes"]["env/observations"]
        assert observation in params_obs["inputs"], 'Observation "%s" must be added, before you can connect it.' % observation
        name, component, cname = source()

        assert converter is not None or source.space_converter, (
            f'"{cname}" does not have a space_converter '
            f'defined under {component} in the spec of "{name}". '
            "Either specify it there, or add an input converter "
            "that acts as a space_converter to this connection."
        )

        # Infer target properties (converter & msg_type) from source
        msg_type_A = get_cls_from_string(source.msg_type)
        output_converter = source.converter.to_dict()
        msg_type_B = get_opposite_msg_cls(msg_type_A, output_converter)
        if converter is None:
            converter = source.space_converter.to_dict()
        msg_type_C = get_opposite_msg_cls(msg_type_B, converter)

        # Set properties in node params of 'env/observations'
        assert len(params_obs["inputs"][observation]) == 0, 'Observation "%s" already connected.' % observation
        msg_type = get_module_type_string(msg_type_C)
        params_obs["inputs"][observation]["msg_type"] = get_module_type_string(msg_type_C)
        mapping = dict(
            msg_type=msg_type,
            delay=0.0,
            window=1,
            skip=False,
            external_rate=None,
            converter=BaseConverter.make("Identity"),
            space_converter=None,
            address=None,
        )
        self._set(params_obs["inputs"], {observation: mapping})
        return converter

    def disconnect(
        self,
        source: GraphView = None,
        target: GraphView = None,
        action: str = None,
        observation: str = None,
        remove: bool = False,
    ):
        """Disconnects a source from a target. The target is reset in self._state to its disconnected state.
        If remove=True, remove observations and actions in the following cases:
        In case of an observation, the complete entry is always removed.
        In case of an action, it is removed if the action is not connected to any other target.
        """
        self._disconnect(source, target, action, observation)
        if remove:
            if action:
                connect_exists = False
                source = ["env/actions", "outputs", action]
                for c in self._state["connects"]:
                    if source == c[0]:
                        connect_exists = True
                        break
                if not connect_exists:
                    self.remove_component(action=action)
            if observation:
                self.remove_component(observation=observation)

    def _disconnect(
        self,
        source: GraphView = None,
        target: GraphView = None,
        action: str = None,
        observation: str = None,
    ):
        """Disconnects a source from a target. The target is reset in self._state to its disconnected state."""
        assert not source or not action, (
            'You cannot specify a source if you wish to disconnect action "%s", '
            "as the action will act as the source." % action
        )
        assert not target or not observation, (
            'You cannot specify a target if you wish to disconnect observation "%s", '
            "as the observation will act as the target." % observation
        )
        assert not (observation and action), (
            "You cannot disconnect an action from an observation," " as such a connection cannot exist."
        )

        # Create source & target entries
        if action:
            assert target is not None, (
                f"If you want to disconnect action {action}, " f"please also specify the corresponding target."
            )

            source = self.get_view("env/actions", ["outputs", action])
        if observation:
            assert source is not None, (
                f"If you want to disconnect observation {observation}," f" please also specify the corresponding source."
            )

            target = self.get_view("env/observations", ["inputs", observation])

        # Check if connection exists
        self._is_selected(self._state, source)
        self._is_selected(self._state, target)

        # Check if connection exists
        connect_exists = False
        idx_connect = None
        for idx, c in enumerate(self._state["connects"]):
            if list(source()) == c[0] and list(target()) == c[1]:
                connect_exists = True
                idx_connect = idx
                break
        assert connect_exists, (
            f"The connection with source={source()} and target={target()} cannot be removed," " because it does not exist."
        )

        # Pop the connection from the state
        self._state["connects"].pop(idx_connect)

        # Reset source params to disconnected state
        if action:
            self._disconnect_action(action)
        else:
            pass  # Nothing to do here (for now)

        # Reset target params to disconnected state (reset to go back to default yaml), i.e. reset window/delay/skip/converter.
        if observation:
            self._disconnect_observation(observation)
        else:
            target_name, target_comp, target_cname = target()
            target_params = self._state["nodes"][target_name]
            target_params[target_comp][target_cname] = self._state["backup"][target_name][target_comp][target_cname]

    def _disconnect_component(self, entry: GraphView, remove=False):
        """Disconnects all associated connects from self._state.
        **DOES NOT** remove observation entries if they are disconnected.
        **DOES NOT** remove action entries if they are disconnect and the last connection.
        """
        was_connected = False
        name, component, cname = entry()
        for source, target in deepcopy(self._state["connects"]):
            source = self.get_view(source[0], source[1:])
            target = self.get_view(target[0], target[1:])
            self._is_selected(self._state, source)
            self._is_selected(self._state, target)
            source_name, source_comp, source_cname = source()
            target_name, target_comp, target_cname = target()
            if source_name == "env/actions":
                action = source_cname
                source = None
            else:
                action = None
                source = source
            if target_name == "env/observations":
                observation = target_cname
                target = None
            else:
                observation = None
                target = target
            if name == source_name and component == source_comp and cname == source_cname:
                self.disconnect(source, target, action, observation, remove=remove)
                was_connected = True
            elif name == target_name and component == target_comp and cname == target_cname:
                self.disconnect(source, target, action, observation, remove=remove)
                was_connected = True
        return was_connected

    def _disconnect_action(self, action: str):
        """Returns the action entry back to its disconnected state.
        That is, remove space_converter if it is not connected to any other targets."""
        params_action = self._state["nodes"]["env/actions"]
        assert action in params_action["outputs"], 'Cannot disconnect action "%s", as it does not exist.' % action
        source = ["env/actions", "outputs", action]
        connect_exists = False
        for c in self._state["connects"]:
            if source == c[0]:
                connect_exists = True
                break
        if not connect_exists:
            params_action["outputs"][action] = dict()

    def _disconnect_observation(self, observation: str):
        """Returns the observation entry back to its disconnected state (i.e. empty dict)."""
        params_obs = self._state["nodes"]["env/observations"]
        assert observation in params_obs["inputs"], 'Cannot disconnect observation "%s", as it does not exist.' % observation
        params_obs["inputs"][observation] = dict()

    def rename(
        self,
        new: str,
        action: Optional[str] = None,
        observation: Optional[str] = None,
    ):
        """Renames action/observation if specified."""
        if action:
            assert observation is None, "Cannot supply both an action and observation."
            entry = self.get_view("env/actions", ["outputs", action])
        if observation:
            assert action is None, "Cannot supply both an action and observation."
            assert observation is not None, "Either an action or observation must be supplied."
            entry = self.get_view("env/observations", ["inputs", observation])
        self._rename_component(entry, new_cname=new)

    def _rename_component(self, entry: GraphView, new_cname: str):
        """Renames the component name (cname) of an entity (node/object) in _state['nodes'] and self._state[connects].
        We cannot change names for node/object components, because their python implementation could depend on it.
        Does not work for feedthroughs.
        """
        name, component, old_cname = entry()
        default = self._state["backup"][name]
        params = self._state["nodes"][name]

        # For now, we only support changing action/observation cnames
        assert name in ["env/observations", "env/actions"], (
            f'Cannot change "{old_cname}" of "{name}". Only name ' "changes to observations and actions are supported."
        )
        assert new_cname not in params[component], f'"{new_cname}" already defined in "{name}" under {component}.'

        # Rename cname in params
        # Does not work for outputs with feedthroughs. Then, both outputs and feedthroughs cnames must be changed.
        for d in (params, default):
            if component in d and old_cname in d[component]:
                assert new_cname not in d[component], f'"{new_cname}" already defined in "{name}" under {component}.'
                d[component][new_cname] = d[component].pop(old_cname)
            if component in d["config"] and old_cname in d["config"][component]:
                assert new_cname not in d["config"][component], f'"{new_cname}" already defined in "{name}" under {component}.'
                d["config"][component].remove(old_cname)
                d["config"][component].append(new_cname)

        # Rename cname in all connects
        for source, target in self._state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target

            if source_comp == component and source_cname == old_cname:
                source[2] = new_cname
            if target_comp == component and target_cname == old_cname:
                target[2] = new_cname

    def set(
        self,
        mapping: Any,
        entry: Optional[GraphView] = None,
        action: Optional[str] = None,
        observation: Optional[str] = None,
        parameter: Optional[str] = None,
    ):
        """Sets parameters in self._state, based on the node/object name. If a component and cname are specified, the
        parameter will be set there. Else, the parameter is set under the "config" key.
        For objects, parameters are set under their agnostic definitions of the components (so not bridge specific).
        If a converter is added, we check if the msg_type changes with the new converter. If so, the component is
        disconnected. See _set_converter for more info.
        """
        self._correct_signature(entry, action, observation)
        if action:
            entry = self.get_view("env/actions", ["outputs", action])
        if observation:
            entry = self.get_view("env/observations", ["inputs", observation])
        if parameter is None:
            if isinstance(mapping, GraphView):
                mapping = mapping.to_dict()
            assert isinstance(mapping, dict), "Can only set mappings of type dict. Else, also set 'parameter=<param_name>'."
        else:
            mapping = {parameter: mapping}

        for parameter, value in mapping.items():
            if parameter:
                getattr(entry, parameter)  # Check if parameter exists
            if parameter == "converter":
                if isinstance(value, ConverterSpec):
                    value = value.params
                elif isinstance(value, GraphView):
                    value = value.to_dict()
                self._set_converter(entry, value)
            else:
                t = entry()
                name = t[0]
                if t[1] == "config":
                    assert parameter not in [
                        "sensors",
                        "actuators",
                        "targets",
                        "states",
                        "inputs",
                        "outputs",
                    ], "You cannot modify component parameters with this function. Use _add/remove_component(..) instead."
                    assert parameter not in ["name"], f"You cannot rename '{name}'."
                    assert parameter not in ["msg_type"], f"You cannot modify msg_type '{value}'."
                    p = self._state["nodes"][name]["config"]
                else:
                    name, component, cname = entry()
                    p = self._state["nodes"][name][component][cname]
                self._set(p, {parameter: value})

    def _set_converter(self, entry: GraphView, converter: Dict):
        """Replaces the converter specified for a node's/object's I/O.
        **DOES NOT** remove observation entries if they are disconnected.
        **DOES NOT** remove action entries if they are disconnect and the last connection.
        """
        _ = entry.converter  # Check if parameter exists
        name, component, cname = entry()
        params = self._state["nodes"][name]

        # Check if converted msg_type of old converter is equal to the msg_type of newly specified converter
        if component == "feedthroughs":
            msg_type = get_cls_from_string(params["outputs"][cname]["msg_type"])
        else:
            msg_type = get_cls_from_string(params[component][cname]["msg_type"])
        converter_old = params[component][cname]["converter"]
        msg_type_ros_old = get_opposite_msg_cls(msg_type, converter_old)
        msg_type_ros_new = get_opposite_msg_cls(msg_type, converter)
        if not msg_type_ros_new == msg_type_ros_old:
            was_disconnected = self._disconnect_component(entry)
        else:
            was_disconnected = False

        # If disconnected action/observation, we cannot add converter so raise error.
        assert not (
            was_disconnected and name in ["env/actions", "env/observations"]
        ), 'Cannot change the converter of action/observation "%s", as it changes the msg_type from ' '"%s" to "%s"' % (
            cname,
            msg_type_ros_old,
            msg_type_ros_new,
        )

        # Make sure the converter is a spaceconverter
        if name in ["env/actions", "env/observations"]:
            converter_cls = get_attribute_from_module(converter["converter_type"])
            assert issubclass(converter_cls, SpaceConverter), (
                'Incorrect converter type for "%s". Action/observation converters should always be a subclass of '
                '"%s/%s".' % (cname, SpaceConverter.__module__, SpaceConverter.__name__)
            )

        # Replace converter
        params[component][cname].pop("converter")
        self._set(params[component][cname], {"converter": converter})

    def get(
        self,
        entry: Optional[Union[GraphView, EntitySpec]] = None,
        action: Optional[str] = None,
        observation: Optional[str] = None,
        parameter: Optional[str] = None,
    ):
        if isinstance(entry, EntitySpec):
            name = entry.params["config"]["name"]
            assert name in self._state["nodes"], f" No entity with name '{name}' in graph."
            return self._state["nodes"][name]
        self._correct_signature(entry, action, observation)
        if action:
            entry = self.get_view("env/actions", ["outputs", action])
        if observation:
            entry = self.get_view("env/observations", ["inputs", observation])
        if parameter:
            return getattr(entry, parameter)
        else:
            return entry

    def register(self):
        """Set the addresses in all incoming components.
        Validate the graph.
        Create params that can be uploaded to the ROS param server."""
        # Check if valid graph.
        assert self.is_valid(plot=False), "Graph not valid."

        # Add addresses based on connections
        state = deepcopy(self._state)
        for source, target in state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target

            # For actions & observations, replace default args
            if source_name == "env/actions":
                default = state["nodes"][target_name]["config"]
                context = {"config": default}
                cname_params = state["nodes"][source_name][source_comp][source_cname]
                substitute_args(cname_params, context, only=["config", "ns"])
                address = "%s/%s/%s" % ("environment", source_comp, source_cname)
            elif target_name == "env/observations":
                default = state["nodes"][source_name]["config"]
                context = {"config": default}
                cname_params = state["nodes"][target_name][target_comp][target_cname]
                substitute_args(cname_params, context, only=["config", "ns"])
                address = "%s/%s/%s" % (source_name, source_comp, source_cname)
            else:
                address = "%s/%s/%s" % (source_name, source_comp, source_cname)
            state["nodes"][target_name][target_comp][target_cname]["address"] = address

        # Initialize param objects
        nodes = []
        objects = []
        render = None
        actions = None
        observations = None
        for name, params in state["nodes"].items():
            if "node_type" in params:
                if name == "env/actions":
                    actions = NodeSpec(params)
                elif name == "env/observations":
                    observations = NodeSpec(params)
                elif name == "env/render":
                    render = NodeSpec(params)
                else:
                    nodes.append(NodeSpec(params))
            else:
                objects.append(ObjectSpec(params))

        assert actions, "No action node defined in the graph."
        assert observations, "No observation node defined in the graph."
        return nodes, objects, actions, observations, render

    def render(
        self,
        source: GraphView,
        rate: float,
        converter: Optional[Dict] = None,
        window: Optional[int] = None,
        delay: Optional[float] = None,
        skip: Optional[bool] = None,
        id: str = "Render",
        **kwargs,
    ):
        # Delete old render node from self._state['nodes'] if it exists
        if "env/render" in self._state["nodes"]:
            self.remove("env/render")

        # Add (new) render node to self._state['node']
        render = Node.make(id, rate=rate, **kwargs)
        self.add(render)

        # Create connection
        target = self.get_view("env/render", depth=["inputs", "image"])
        self.connect(
            source=source,
            target=target,
            converter=converter,
            window=window,
            delay=delay,
            skip=skip,
        )

    def save(self, path: str):
        with open(path, "w") as outfile:
            yaml.dump(self._state, outfile, default_flow_style=False)
        pass

    def load(self, path: str):
        with open(path, "r") as stream:
            try:
                self._state = yaml.safe_load(stream)
                # self._state = yaml.load(path)
            except yaml.YAMLError as exc:
                print(exc)

    def gui(self):
        try:
            from eagerx_gui import launch_gui
        except ImportError as e:
            rospy.logwarn(
                f"{e}. You will likely have to install eagerx-gui. Please visit the eagerx_packages repository for installation instructions."
            )
            return

        self._state = launch_gui(deepcopy(self._state))

    @staticmethod
    def _is_selected(state: Dict, entry: GraphView):
        """Check if provided entry was selected in params."""
        name, component, cname = entry()
        params = state["nodes"][name]
        component = "outputs" if component == "feedthroughs" else component
        assert cname in params["config"][component], f'"{cname}" not selected in "{name}" under "config" in {component}.'

    @staticmethod
    def _correct_signature(
        entry: Optional[GraphView] = None,
        action: Optional[str] = None,
        observation: Optional[str] = None,
    ):
        # assert only action, only observation, or only entry
        if entry:  # component parameter
            assert action is None, "If 'entry' is specified, action argument cannot be specified."
            assert observation is None, "If 'entry' is specified, observation argument cannot be specified."
        if action:
            assert observation is None, "If action is specified, observation must be None."
            assert entry is None, "If action is specified, the 'entry' argument cannot be specified."
        if observation:
            assert action is None, "If observation is specified, action must be None."
            assert entry is None, "If action is specified, the 'entry' argument cannot be specified."

    def is_valid(self, plot=True):
        return self._is_valid(self._state, plot=plot)

    @staticmethod
    def _is_valid(state, plot=True):
        state = deepcopy(state)
        Graph.check_msg_types_are_consistent(state)
        Graph.check_inputs_have_address(state)
        Graph.check_graph_is_acyclic(state, plot=plot)
        # Graph.check_exists_compatible_bridge(state)
        return True

    @staticmethod
    def check_msg_type(source, target, state):
        # Convert the source msg_type to target msg_type with converters:
        # msg_type_source --> output_converter --> msg_type_ROS --> input_converter --> msg_type_target
        msg_type_out = get_cls_from_string(source.msg_type)
        converter_out = source.converter.to_dict()
        msg_type_ros = get_opposite_msg_cls(msg_type_out, converter_out)
        converter_in = target.converter.to_dict()
        msg_type_in = get_opposite_msg_cls(msg_type_ros, converter_in)

        # Verify that this msg_type_in is the same as the msg_type specified in the target
        target_name, target_comp, target_cname = target()
        if target_comp == "feedthroughs":
            msg_type_in_registered = get_cls_from_string(state["nodes"][target_name]["outputs"][target_cname]["msg_type"])
        else:
            msg_type_in_registered = get_cls_from_string(target.msg_type)

        msg_type_str = msg_type_error(
            source(),
            target(),
            msg_type_out,
            converter_out,
            msg_type_ros,
            converter_in,
            msg_type_in,
            msg_type_in_registered,
        )
        assert msg_type_in == msg_type_in_registered, msg_type_str

    @staticmethod
    def check_msg_types_are_consistent(state):
        for source, target in state["connects"]:
            source = Graph._get_view(state, source[0], source[1:])
            target = Graph._get_view(state, target[0], target[1:])
            Graph.check_msg_type(source, target, state)
        return True

    @staticmethod
    def check_inputs_have_address(state):
        state = deepcopy(state)
        for source, target in state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target
            address = "%s/%s/%s" % (source_name, source_comp, source_cname)
            state["nodes"][target_name][target_comp][target_cname]["address"] = address

        for name, params in state["nodes"].items():
            if "node_type" in params:
                for component in params["config"]:
                    if component not in [
                        "inputs",
                        "outputs",
                        "targets",
                        "feedthroughs",
                        "states",
                    ]:
                        continue
                    for cname in params["config"][component]:
                        assert cname in params[component], (
                            f'"{cname}" was selected in {component} of "{name}", ' "but has no implementation."
                        )
                        if component not in ["inputs", "targets", "feedthroughs"]:
                            continue
                        p = params[component][cname]
                        assert "address" in p and p["address"] is not None, (
                            f'"{cname}" was selected in {component} '
                            f'of "{name}", but no address could be '
                            "produced/inferred. Either deselect it, "
                            "or connect it."
                        )
            else:
                for component in params["config"]:
                    if component not in ["sensors", "actuators", "states"]:
                        continue
                    for cname in params["config"][component]:
                        assert cname in params[component], (
                            f'"{cname}" was selected in {component} of "{name}", ' "but has no (agnostic) implementation."
                        )
                        if component not in ["actuators"]:
                            continue
                        assert "address" in params[component][cname], (
                            f'"{cname}" was selected in {component} of "{name}", '
                            "but no address could be produced/inferred. "
                            "Either deselect it, or connect it."
                        )
        return True

    @staticmethod
    def check_graph_is_acyclic(state, plot=True):
        # Add nodes
        G = nx.MultiDiGraph()
        label_mapping = {
            "env/observations/set": "observations",
            "env/render/done": "render",
        }
        for node, params in state["nodes"].items():
            default = params["config"]
            if "node_type" not in state["nodes"][node]:  # Object
                if "sensors" in default:
                    for cname in default["sensors"]:
                        name = "%s/sensors/%s" % (node, cname)
                        G.add_node(
                            "%s/sensors/%s" % (node, cname),
                            remain_active=False,
                            always_active=True,
                            is_stale=False,
                            has_tick=False,
                        )
                        if not ("actuators" in default and cname in default["actuators"]):
                            label_mapping[name] = "%s/%s" % (node, cname)
                if "actuators" in default:
                    for cname in default["actuators"]:
                        name = "%s/actuators/%s" % (node, cname)
                        G.add_node(
                            "%s/actuators/%s" % (node, cname),
                            remain_active=True,
                            always_active=False,
                            is_stale=False,
                            has_tick=False,
                        )
                        if not ("sensors" in default and cname in default["sensors"]):
                            label_mapping[name] = "%s/%s" % (node, cname)
            else:  # node
                if "outputs" in default:
                    for cname in default["outputs"]:
                        name = "%s/%s" % (node, cname)
                        if name == "env/actions/set":
                            continue
                        G.add_node(
                            name,
                            remain_active=False,
                            always_active=False,
                            is_stale=False,
                            has_tick=False,
                        )

        # Add edges
        for cname in state["nodes"]["env/actions"]["config"]["outputs"]:
            if cname == "set":
                continue
            name = "env/actions/%s" % cname
            label_mapping[name] = "actions/%s" % cname
            G.add_edge(
                "env/observations/set",
                name,  # key='%s/%s' % ('inputs', 'observations_set'),
                feedthrough=False,
                style="solid",
                color="black",
                alpha=1.0,
                is_stale=False,
                skip=False,
                source=("env/observations", "outputs", "set"),
                target=("env/actions", "inputs", "observations_set"),
            )
        target_comps = ["inputs", "actuators", "feedthroughs"]
        source_comps = ["outputs", "sensors"]
        for source, target in state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target
            if source_comp in source_comps and target_comp in target_comps:
                # Determine source node name
                if "node_type" in state["nodes"][source_name]:
                    source_edge = "%s/%s" % (source_name, source_cname)
                else:
                    source_edge = "%s/%s/%s" % (source_name, source_comp, source_cname)
                # Determine target node name
                target_edges = []
                target_default = state["nodes"][target_name]["config"]
                if "node_type" in state["nodes"][target_name]:
                    for cname in target_default["outputs"]:
                        target_edge = "%s/%s" % (target_name, cname)
                        target_edges.append(target_edge)
                else:
                    target_edge = "%s/%s/%s" % (target_name, target_comp, target_cname)
                    target_edges.append(target_edge)

                # Determine stale nodes in real_reset routine via feedthrough edges
                if target_comp == "feedthroughs":
                    feedthrough = True
                else:
                    feedthrough = False

                # Determine edges that do not break DAG property (i.e. edges that are skipped)
                skip = state["nodes"][target_name][target_comp][target_cname]["skip"]
                color = "green" if skip else "black"
                style = "dotted" if skip else "solid"

                # Add edge
                for target_edge in target_edges:
                    G.add_edge(
                        source_edge,
                        target_edge,
                        color=color,
                        feedthrough=feedthrough,
                        style=style,
                        alpha=1.0,
                        is_stale=False,
                        skip=skip,
                        source=source,
                        target=target,
                    )

        # Color nodes based on in/out going edges
        not_active = is_stale(G)
        color_nodes(G)
        color_edges(G)

        # Remap action & observation labels to more readable form
        G = nx.relabel_nodes(G, label_mapping)

        # Check if graph is acyclic (excluding 'skip' edges)
        H, cycles = episode_graph(G)
        is_dag = nx.is_directed_acyclic_graph(H)

        # Plot graphs
        if plot:
            fig_env, ax_env = plt.subplots(nrows=1, ncols=1)
            ax_env.set_title("Communication graph (episode)")
            _, _, _, pos = plot_graph(G, k=2, ax=ax_env)
            plt.show()

        # Assert if graph is a directed-acyclical graph (DAG)
        cycle_strs = ["Algebraic loops detected: "]
        for idx, connect in enumerate(cycles):
            connect.append(connect[0])
            s = " Loop %s: " % idx
            n = "\n" + "".join([" "] * len(s)) + "...-->"
            s = "\n\n" + s + "...-->"
            for idx in range(len(connect) - 1):
                tmp, target = connect[idx]
                source, tmp2 = connect[idx + 1]
                source_name, source_comp, source_cname = source
                target_name, target_comp, target_cname = target
                assert source_name == target_name, "Source and target not equal: %s, %s" % (source, target)
                connect_name = "%s/%s/%s][%s/%s/%s" % tuple(list(source) + list(target))
                node_name = ("Node: " + source_name).center(len(connect_name), " ")
                s += "[%s]-->" % connect_name
                n += "[%s]-->" % node_name
            s += "..."
            n += "..."
            cycle_strs.append(s)
            cycle_strs.append(n)
            connect.pop(-1)
        assert is_dag, "".join(cycle_strs)
        assert len(not_active) == 0, (
            'Stale episode graph detected. Nodes "%s" will be stale, while they must be active (i.e. connected) in order for the graph to resolve (i.e. not deadlock).'
            % not_active
        )

        # Create a shallow copy graph that excludes feedthrough edges
        F = reset_graph(G)
        not_active = is_stale(F)
        color_nodes(F)
        color_edges(F)

        # Plot graphs
        if plot:
            fig_reset, ax_reset = plt.subplots(nrows=1, ncols=1)
            ax_reset.set_title("Communication graph (reset)")
            _, _, _, pos = plot_graph(F, pos=pos, ax=ax_reset)
            plt.show()

        # Assert if reset graph is not stale
        has_real_reset = len([e for e, ft in nx.get_edge_attributes(G, "feedthrough").items() if ft]) > 0
        assert len(not_active) == 0 or not has_real_reset, (
            'Stale reset graph detected. Nodes "%s" will be stale, while they must be active (i.e. connected) in order for the graph to resolve (i.e. not deadlock).'
            % not_active
        )
        return True

    @staticmethod
    def check_exists_compatible_bridge(state, tablefmt="fancy_grid"):
        msg = "Bridge implementations are not part of the engine graph anymore."
        raise NotImplementedError(msg)
        # # Bridges are headers
        # from tabulate import tabulate
        # bridges = []
        # objects = []
        # for node, params in state["nodes"].items():
        #     default = params["config"]
        #     # todo: only check bridge if it concerns one of the selected components
        #     if "node_type" not in state["nodes"][node]:  # Object
        #         params = state["nodes"][node]
        #         entity_id = params["config"]["entity_id"]
        #         obj_name = params["config"]["name"]
        #         entry = [obj_name, entity_id]
        #
        #         # Add all (unknown) bridges to the list
        #         for key, _value in params.items():
        #             if key in [
        #                 "entity_type",
        #                 "config",
        #                 "sensors",
        #                 "actuators",
        #                 "states",
        #             ]:
        #                 continue
        #             if key not in bridges:
        #                 bridges.append(key)
        #
        #         # See what bridges support all object components
        #         for b in bridges:
        #             if b in params:
        #                 for component in ["sensors", "actuators", "states"]:
        #                     if component in default:
        #                         for cname in default[component]:
        #                             if component in params[b] and cname in params[b][component]:
        #                                 e_str = "x"  # Component entry is supported
        #                             else:
        #                                 e_str = " "  # Component entry is not supported
        #                                 break  # Break if entry in component is not supported
        #             else:  # Bridge name not even mentioned in object config
        #                 e_str = " "
        #             entry.append(e_str)
        #         objects.append(entry)
        #
        # # Fill up incompatible bridges that were added after object entries
        # for entry in objects:
        #     for _ in bridges[len(entry) - 2 :]:
        #         entry.append(" ")
        #
        # # Get compatible bridges
        # compatible = []
        # for idx, b in enumerate(bridges):
        #     idx = idx + 2
        #     for entry in objects:
        #         c = [True if entry[idx] == "x" else False for entry in objects]
        #     if len(c) == len(objects):
        #         compatible.append(b)
        #
        # # Objects are entries
        # headers = ["name", "object"]
        # for b in bridges:
        #     headers.append(b.replace("/", "/\n"))
        #
        # # Assert if there are compatible bridges
        # tabulate_str = tabulate(
        #     objects,
        #     headers=headers,
        #     tablefmt=tablefmt,
        #     colalign=["center"] * len(headers),
        # )
        # assert len(compatible), (
        #     "No compatible bridges for the selected objects. Ensure that all components, selected in each object, is supported by a common bridge.\n%s"
        #     % tabulate_str
        # )
        # return tabulate_str

    @staticmethod
    @supported_types(str, int, list, float, bool, dict, EntitySpec, GraphView, None)
    def _set(state, mapping):
        merge(state, mapping)

    @staticmethod
    def _get_view(state, name: str, depth: Optional[List[str]] = None):
        depth = depth if depth else []
        depth = [name] + depth
        return GraphView(Graph(state), depth=depth, name=name)

    def get_view(self, name: str, depth: Optional[List[str]] = None):
        depth = depth if depth else []
        depth = [name] + depth
        return GraphView(self, depth=depth, name=name)



from typing import Optional
from scipy.signal import butter, sosfilt

# IMPORT ROS
from std_msgs.msg import Float32MultiArray

# IMPORT EAGERX
import eagerx.core.register as register
from eagerx.utils.utils import Msg
from eagerx.core.entities import Node, Processor, SpaceConverter
from eagerx.core.constants import process


class ButterworthFilter(Node):
    @staticmethod
    @register.spec("ButterworthFilter", Node)
    def spec(
        spec,
        name: str,
        rate: float,
        index: int = 0,
        N: int = 2,
        Wn: float = 1,
        btype: str = "lowpass",
        process: Optional[int] = process.NEW_PROCESS,
        color: Optional[str] = "grey",
    ):
        """
        Butterworth filter implementation based on scipy.signal/butter, scipy.signal/sosfilt.

        :param spec: Not provided by user.
        :param name: Node name
        :param rate: Rate at which callback is called.
        :param index: Index (related to Float32MultiArray.data[index])
        :param N: The order of the filter
        :param Wn: The critical frequency or frequencies
        :param btype: {'lowpass', 'highpass', 'bandpass', 'bandstop'}
        :param process: {0: NEW_PROCESS, 1: ENVIRONMENT, 2: BRIDGE, 3: EXTERNAL}
        :param color: console color of logged messages. {'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', 'grey'}
        :return:
        """
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(ButterworthFilter)

        # Modify default node params
        spec.config.name = name
        spec.config.rate = rate
        spec.config.process = process
        spec.config.color = color
        spec.config.inputs = ["signal"]
        spec.config.outputs = ["filtered"]

        # Modify custom node params
        spec.config.N = N
        spec.config.Wn = Wn
        spec.config.btype = btype

        # Add converter & space_converter
        spec.inputs.signal.window = "$(config N)"
        spec.inputs.signal.converter = Processor.make("GetIndex_Float32MultiArray", index=index)
        spec.inputs.signal.space_converter = SpaceConverter.make("Space_Float32MultiArray", [-3], [3], dtype="float32")

    def initialize(self, N, Wn, btype):
        for i in self.inputs:
            if i["name"] == "signal":
                assert (
                    int(i["window"]) >= N
                ), "The window size of the signal {} is too small to create a filter with order {}.".format(i["window"], N)
        self.filter = butter(N, Wn, btype, output="sos", fs=self.rate)
        self.N = N

    @register.states()
    def reset(self):
        pass

    @register.inputs(signal=Float32MultiArray)
    @register.outputs(filtered=Float32MultiArray)
    def callback(self, t_n: float, signal: Optional[Msg] = None):
        msgs = signal.msgs
        if len(msgs) >= self.N:
            unfiltered = [msgs[i].data[0] for i in range(-self.N, 0)]
            filtered = msgs[-1].data if None in unfiltered else [sosfilt(self.filter, unfiltered)[-1]]
        elif len(msgs) > 0:
            filtered = msgs[-1].data
        else:
            filtered = [0.0]
        return dict(filtered=Float32MultiArray(data=filtered))



# OTHER IMPORTS
from typing import Optional, List
from math import isclose

# ROS IMPORTS
import rospy
from std_msgs.msg import UInt64

# EAGERx IMPORTS
from eagerx.core.constants import process, ERROR
from eagerx.core.entities import Bridge, SpaceConverter
import eagerx.core.register as register
from eagerx.utils.utils import Msg


class TestBridgeNode(Bridge):
    def initialize(self, num_substeps: int, nonreactive_address: str):
        # Initialize any simulator here, that is passed as reference to each simnode
        self.simulator = None

        # If real_time bridge, assert that real_time_factor == 1 & is_reactive=False.

        # Initialize nonreactive input (Only required for this test bridge implementation
        self.nonreactive_pub = rospy.Publisher(self.ns + nonreactive_address, UInt64, queue_size=0, latch=True)

    @staticmethod
    @register.spec("TestBridge", Bridge)
    def spec(
        spec,
        rate,
        process: Optional[int] = process.NEW_PROCESS,
        is_reactive: Optional[bool] = True,
        real_time_factor: Optional[float] = 0,
        simulate_delays: Optional[bool] = True,
        log_level: Optional[int] = ERROR,
        states: Optional[List[str]] = None,
    ):
        """TestBridge spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        TestBridgeNode.initialize_spec(spec)

        # Modify default bridge params
        spec.config.rate = rate
        spec.config.process = process
        spec.config.is_reactive = is_reactive
        spec.config.real_time_factor = real_time_factor
        spec.config.simulate_delays = simulate_delays
        spec.config.log_level = log_level
        spec.config.color = "magenta"
        spec.config.states = states if states else ["param_1"]

        # Add custom params
        spec.config.num_substeps = 10
        spec.config.nonreactive_address = "/nonreactive_input_topic"  # Only required to test nonreactive inputs

        # Add state: "param_1"
        spec.states.param_1.space_converter = SpaceConverter.make("Space_RosUInt64", low=[0], high=[100], dtype="uint64")

    @register.bridge_config(req_arg=None, xacro="$(find some_package)/urdf/object.urdf.xacro")
    def add_object(self, config, bridge_config, node_params, state_params):
        # add object to simulator (we have a ref to the simulator with self.simulator)
        rospy.loginfo(f'Adding object "{config["name"]}" of type "{config["entity_id"]}" to the simulator.')

    def pre_reset(self, param_1: Optional[UInt64] = None):
        return "PRE RESET RETURN VALUE"

    @register.states(param_1=UInt64)
    def reset(self, param_1: Optional[UInt64] = None):
        # Publish nonreactive input (this is only required for simulation setup)
        self.nonreactive_pub.publish(UInt64(data=0))
        return "POST RESET RETURN VALUE"

    @register.outputs(tick=UInt64)
    def callback(self, t_n: float, **kwargs: Optional[Msg]):
        # Publish nonreactive input
        self.nonreactive_pub.publish(UInt64(data=self.num_ticks))

        # Verify that # of ticks equals internal counter
        node_tick = t_n * self.rate
        if not isclose(self.num_ticks, node_tick):
            rospy.logerr(
                f"[{self.name}][callback]: ticks not equal (self.num_ticks={self.num_ticks}, node_tick={round(node_tick)})."
            )

        # Verify that all timestamps are smaller or equal to node time
        t_n = node_tick * (1 / self.rate)
        for i in self.inputs:
            name = i["name"]
            if name in kwargs:
                t_i = kwargs[name].info.t_in
                if len(t_i) > 0 and not all((t.sim_stamp - t_n) <= 1e-7 for t in t_i if t is not None):
                    rospy.logerr(f"[{self.name}][{name}]: Not all t_i are smaller or equal to t_n.")



# OTHER IMPORTS
from typing import Optional, List

# ROS IMPORTS
from std_msgs.msg import UInt64, String

# EAGERx IMPORTS
from eagerx.bridges.test.bridge import TestBridgeNode
from eagerx.core.entities import Object, EngineNode, SpaceConverter, EngineState, BaseConverter, Converter, Processor
from eagerx.core.specs import ObjectSpec
from eagerx.core.graph_engine import EngineGraph
import eagerx.core.register as register


class Arm(Object):
    entity_id = "Arm"

    @staticmethod
    @register.sensors(N6=UInt64, N7=UInt64)
    @register.actuators(N8=String, ref_vel=UInt64, N12=UInt64)
    @register.simstates(N9=UInt64, N10=UInt64)
    @register.config(position=[0, 0, 0], orientation=[0, 0, 0], low=None, string=None, test_string=None, test_list=None)
    def agnostic(spec: ObjectSpec, rate):
        """Agnostic definition of the Arm object"""
        # Set state properties: space_converters
        spec.sensors.N6.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.sensors.N7.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.sensors.N6.rate = rate
        spec.sensors.N7.rate = 2

        # Set actuator properties: space_converters
        spec.actuators.N8.space_converter = SpaceConverter.make("Space_RosString", [0], [100], dtype="uint64")
        # spec.actuators.N12.space_converter = SpaceConverter.make("Space_RosString", [0], [100], dtype="uint64")
        spec.actuators.N12.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.actuators.ref_vel.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.actuators.N8.rate = rate
        spec.actuators.N12.rate = rate
        spec.actuators.ref_vel.rate = 1

        # Set state properties: space_converters
        spec.states.N9.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")
        spec.states.N10.space_converter = SpaceConverter.make("Space_RosUInt64", [0], [100], dtype="uint64")

        # Test AgnosticSpec
        spec.actuators.N8.rate = spec.actuators.N8.rate
        spec.actuators.N8 = spec.actuators.N8

    @staticmethod
    @register.spec(entity_id, Object)
    def spec(
        spec: ObjectSpec,
        name: str = None,
        sensors: Optional[List[str]] = None,
        actuators: Optional[List[str]] = None,
        states: Optional[List[str]] = None,
        position: Optional[List[str]] = None,
        orientation: Optional[List[str]] = None,
        string: Optional[str] = "test_arg",
        test_string: Optional[str] = "$(config string)",
        test_list: Optional[str] = "$(config orientation)",
        low: Optional[int] = 0,
    ):
        """Object spec of Arm"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        Arm.initialize_spec(spec)

        # Modify default agnostic params
        # Only allow changes to the agnostic params (rates, windows, (space)converters, etc...
        spec.config.name = name
        spec.config.sensors = sensors if sensors else ["N6", "N7"]
        spec.config.actuators = actuators if actuators else ["ref_vel"]
        spec.config.states = states if states else ["N9", "N10"]

        # Change custom params
        spec.config.position = position if position else [0, 0, 0]
        spec.config.orientation = orientation if orientation else [0, 0, 0]
        spec.config.string = string
        spec.config.test_string = test_string
        spec.config.test_list = test_list
        spec.config.low = low

        # Test ObjectSpec
        spec.sensors.N6 = spec.sensors.N6
        spec.config.name = spec.config.name

        # Add agnostic definition
        Arm.agnostic(spec, rate=15)

    @staticmethod
    @register.bridge(entity_id, TestBridgeNode)
    def test_bridge(spec: ObjectSpec, graph: EngineGraph):
        """Engine-specific implementation of the Arm with the test bridge."""
        # Set bridge_config
        spec.TestBridge.req_arg = "TEST"
        spec.TestBridge.xacro = "$(find some_package)/urdf/arm.urdf.xacro"

        # Create simstates
        spec.TestBridge.states.N9 = EngineState.make("TestEngineState", test_arg="arg_N9")

        # Create sensor engine nodes
        N6 = EngineNode.make(
            "TestSensor",
            "N6",
            rate=spec.sensors.N6.rate,
            process=2,
            inputs=["tick", "in_1"],
            outputs=["out_1"],
            states=["state_1"],
            test_arg=spec.TestBridge.req_arg,
        )
        N7 = EngineNode.make(
            "TestSensor",
            "N7",
            rate=spec.sensors.N7.rate,
            process=2,
            inputs=["tick", "in_1"],
            outputs=["out_1"],
            states=[],
            test_arg=spec.config.test_string,
        )

        # Create actuator engine nodes
        N8 = EngineNode.make(
            "TestActuator",
            "N8",
            rate=spec.actuators.N8.rate,
            process=2,
            inputs=["tick", "in_2", "in_3"],
            outputs=["out_1"],
            test_arg=spec.config.test_string,
            color="green",
        )
        ref_vel = EngineNode.make(
            "TestActuator",
            "ref_vel",
            rate=spec.actuators.ref_vel.rate,
            process=2,
            inputs=["tick", "in_1", "in_2"],
            outputs=["out_1"],
            test_arg=spec.config.test_string,
            color="green",
        )

        # Test SpecificSpec:
        _ = spec.__str__()
        spec.TestBridge.states.N9.test_arg = "test2"

        # Test EngineGraph: Add/remove sensor
        graph.add(N6)
        _ = graph.__str__()
        graph.add_component(N6.outputs.out_2)
        graph.remove_component(N6.outputs.out_2)
        graph.connect(source=N6.outputs.out_1, sensor="N6")
        graph.remove(N6)
        N6.set_graph(None)

        # Test EngineGraph: Add/remove actuator
        graph.add(N8)
        graph.connect(actuator="N8", target=N8.inputs.in_3)
        graph.remove("N8")
        N8.set_graph(None)

        # Test EngineGraph: Remove component
        graph.add(N6)
        graph.connect(source=N6.outputs.out_1, sensor="N6")
        graph.remove_component(N6.outputs.out_1)
        graph.add_component(N6.outputs.out_1)
        graph.remove("N6")
        N6.set_graph(None)

        # Test EngineGraph: Remove component
        graph.add(N8)
        graph.connect(actuator="N8", target=N8.inputs.in_3)
        graph.remove_component(N8.inputs.in_3)
        graph.add_component(N8.inputs.in_3)
        graph.remove("N8")
        N8.set_graph(None)

        # Test EngineGraph: parameters
        graph.add([N6, N8])
        graph.connect(source=N6.outputs.out_1, sensor="N6")
        graph.connect(actuator="N8", target=N8.inputs.in_3)
        graph.get(N6.outputs.out_1.converter)
        graph.set({"test_arg": "NEW_ARG"}, N6.config)
        graph.get(actuator="N8", parameter="converter")
        graph.get(sensor="N6", parameter="converter")
        graph.get(sensor="N6")
        graph.get(actuator="N8")
        graph.get(N6.outputs.out_1)
        graph.set(graph.get(N6.outputs.out_1), N6.outputs.out_1)
        graph.get(N6)
        graph.remove(["N6", "N8"])
        N6.set_graph(None)
        N8.set_graph(None)

        # Add nodes to graph and connect them to actuators/sensors
        graph.add([N6, N7, N8, ref_vel])

        # Connect sensors & actuators to engine nodes
        graph.connect(source=N6.outputs.out_1, sensor="N6")
        graph.disconnect(source=N6.outputs.out_1, sensor="N6")
        graph.connect(source=N6.outputs.out_1, sensor="N6")
        graph.connect(source=N7.outputs.out_1, sensor="N7")
        graph.connect(actuator="N8", target=N8.inputs.in_3)
        graph.connect(actuator="ref_vel", target=ref_vel.inputs.in_1)

        # Interconnect engine nodes
        id = BaseConverter.make("Identity")
        graph.connect(
            source=N8.outputs.out_1,
            target=N7.inputs.in_1,
            skip=True,
        )
        graph.connect(source=N6.outputs.out_1, target=N8.inputs.in_2, window=1, converter=id)
        graph.connect(source=N6.outputs.out_1, target=ref_vel.inputs.in_2, delay=0.0)

        # Connect simnode with external address (cannot be actuator or sensor)
        graph.connect(
            address="$(ns env_name)/nonreactive_input_topic",
            target=N6.inputs.in_1,
            external_rate=20,
        )
        graph.disconnect(target=N6.inputs.in_1)
        graph.connect(
            address="$(ns env_name)/nonreactive_input_topic",
            target=N6.inputs.in_1,
            external_rate=20,
        )

        # Test EngineGraph: Test acyclic check
        import matplotlib.pyplot as plt

        plt.ion()

        graph.add_component(N8.inputs.in_1)
        graph.connect(source=N7.outputs.out_1, target=N8.inputs.in_1)
        graph.disconnect(source=N8.outputs.out_1, target=N7.inputs.in_1)
        graph.connect(source=N8.outputs.out_1, target=N7.inputs.in_1)
        try:
            graph.is_valid(plot=True)
        except AssertionError as e:
            if "Algebraic" in e.args[0]:
                pass
            else:
                raise
        # Reconnect
        graph.remove_component(N8.inputs.in_1)
        graph.disconnect(source=N8.outputs.out_1, target=N7.inputs.in_1)
        graph.connect(source=N8.outputs.out_1, target=N7.inputs.in_1, skip=True)

        # Connect multiple engine nodes to same actuator
        N11 = EngineNode.make(
            "TestActuator",
            "N11",
            rate=spec.actuators.N8.rate,
            process=2,
            inputs=["tick", "in_3"],
            outputs=["out_1"],
            test_arg=spec.config.test_string,
            color="green",
        )
        graph.add(N11)
        graph.connect(actuator="N8", target=N11.inputs.in_3)


class Viper(Arm):
    entity_id = "Viper"

    @staticmethod
    @register.spec(entity_id, Object)
    def spec(
        spec: ObjectSpec,
        name: str,
        sensors: Optional[List[str]] = None,
        actuators: Optional[List[str]] = None,
        states: Optional[List[str]] = None,
        position: Optional[List[float]] = None,
        orientation: Optional[List[float]] = None,
        string: Optional[str] = "test_arg",
        test_string: Optional[str] = "$(config string)",
        test_list: Optional[str] = "$(config orientation)",
        low: Optional[int] = 0,
    ):
        """Object spec of Viper"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        Viper.initialize_spec(spec)

        # Modify default agnostic params
        # Only allow changes to the agnostic params (rates, windows, (space)converters, etc...
        spec.config.name = name
        spec.config.sensors = sensors if sensors else ["N6", "N7"]
        spec.config.actuators = actuators if actuators else ["ref_vel"]
        spec.config.states = states if states else ["N9", "N10"]

        # Change custom params
        spec.config.position = position if position else [0, 0, 0]
        spec.config.orientation = orientation if orientation else [0, 0, 0]
        spec.config.string = string
        spec.config.test_string = test_string
        spec.config.test_list = test_list
        spec.config.low = low

        # Add agnostic definition
        Viper.agnostic(spec, rate=15)

    @staticmethod
    @register.bridge(entity_id, TestBridgeNode)
    def test_bridge(spec: ObjectSpec, graph: EngineGraph):
        """Engine-specific implementation of the Viper with the test bridge."""
        # Set object arguments
        spec.TestBridge.req_arg = "TEST ARGUMENT"
        spec.TestBridge.xacro = "$(find some_package)/urdf/viper.urdf.xacro"

        # Create simstates
        spec.TestBridge.states.N9 = EngineState.make("TestEngineState", test_arg="arg_N9")
        spec.TestBridge.states.N10 = EngineState.make("TestEngineState", test_arg="arg_N10")

        # Create sensor engine nodes
        N6 = EngineNode.make(
            "TestSensor",
            "N6",
            rate=spec.sensors.N6.rate,
            process=2,
            inputs=["tick", "in_1"],
            outputs=["out_1"],
            states=["state_1"],
            test_arg=spec.TestBridge.req_arg,
        )
        N7 = EngineNode.make(
            "TestSensor",
            "N7",
            rate=spec.sensors.N7.rate,
            process=2,
            inputs=["tick", "in_1"],
            outputs=["out_1"],
            states=[],
            test_arg=spec.config.test_string,
        )

        # Create actuator engine nodes
        N8 = EngineNode.make(
            "TestActuator",
            "N8",
            rate=spec.actuators.N8.rate,
            process=2,
            inputs=["tick", "in_2", "in_3"],
            outputs=["out_1"],
            test_arg=spec.config.test_string,
            color="green",
        )
        ref_vel = EngineNode.make(
            "TestActuator",
            "ref_vel",
            rate=spec.actuators.ref_vel.rate,
            process=2,
            inputs=["tick", "in_1", "in_2"],
            outputs=["out_1"],
            test_arg=spec.config.test_string,
            color="green",
        )

        # Add IdentityProcessor to N6
        IdentityProcessor = Processor.make("IdentityProcessor")
        N6.outputs.out_1.converter = IdentityProcessor

        # Add nodes to graph and connect them to actuators/sensors
        graph.add([N6, N7, N8, ref_vel])

        # Connect sensors & actuators to engine nodes
        graph.connect(source=N6.outputs.out_1, sensor="N6")
        graph.disconnect(source=N6.outputs.out_1, sensor="N6")
        graph.connect(source=N6.outputs.out_1, sensor="N6")
        graph.connect(source=N7.outputs.out_1, sensor="N7")
        graph.connect(actuator="N8", target=N8.inputs.in_3)
        graph.connect(actuator="ref_vel", target=ref_vel.inputs.in_1)

        # Interconnect engine nodes
        graph.connect(
            source=N8.outputs.out_1,
            target=N7.inputs.in_1,
            skip=True,
        )
        graph.connect(source=N6.outputs.out_1, target=N8.inputs.in_2)
        graph.connect(source=N6.outputs.out_1, target=ref_vel.inputs.in_2)

        # Connect simnode with external address (cannot be actuator or sensor)
        graph.connect(
            address="$(ns env_name)/nonreactive_input_topic",
            target=N6.inputs.in_1,
            external_rate=20,
        )
        graph.disconnect(target=N6.inputs.in_1)
        graph.connect(
            address="$(ns env_name)/nonreactive_input_topic",
            target=N6.inputs.in_1,
            external_rate=20,
        )

        # Test Adding converters to actuators
        N12 = EngineNode.make(
            "TestActuator",
            "N12",
            rate=spec.actuators.N12.rate,
            process=2,
            inputs=["tick", "in_3"],
            outputs=["out_1"],
            test_arg=spec.config.test_string,
            color="green",
        )
        graph.add(N12)
        RosString_RosUInt64 = Converter.make("RosString_RosUInt64", test_arg="test")
        graph.connect(actuator="N12", target=N12.inputs.in_3, converter=RosString_RosUInt64)



# ROS SPECIFIC
import rospy
import rosparam
from std_msgs.msg import UInt64

# RxEAGER
from eagerx.utils.utils import substitute_args
from eagerx.core.constants import process, log, log_levels_ROS

# OTHER
import atexit
import importlib
import subprocess
from time import sleep
from typing import List, Dict, Union, Any
from functools import partial, wraps


@wraps(rospy.init_node)
def initialize(*args, log_level=log.INFO, **kwargs):
    roscore = launch_roscore()  # First launch roscore (if not already running)
    try:
        rospy.init_node(*args, log_level=log_levels_ROS[log_level], **kwargs)
    except rospy.exceptions.ROSException as e:
        rospy.logwarn(e)
    if roscore:
        atexit.register(roscore.shutdown)
    return roscore


def launch_roscore():
    import roslaunch

    uuid = roslaunch.rlutil.get_or_generate_uuid(options_runid=None, options_wait_for_master=False)
    roslaunch.configure_logging(uuid)
    roscore = roslaunch.parent.ROSLaunchParent(uuid, roslaunch_files=[], is_core=True)

    try:
        roscore.start()
    except roslaunch.core.RLException:
        rospy.logwarn(
            "Roscore cannot run as another roscore/master is already running. Continuing without re-initializing the roscore."
        )
        roscore = None
    return roscore


def launch_node(launch_file, args):
    import roslaunch

    cli_args = [substitute_args(launch_file)] + args
    roslaunch_args = cli_args[1:]
    roslaunch_file = [(roslaunch.rlutil.resolve_launch_arguments(cli_args)[0], roslaunch_args)]
    uuid = roslaunch.rlutil.get_or_generate_uuid(None, False)
    # roslaunch.configure_logging(uuid)  # THIS RESETS the log level. Can we do without this line? Are ROS logs stil being made?
    launch = roslaunch.parent.ROSLaunchParent(uuid, roslaunch_file)
    return launch


def launch_node_as_subprocess(executable, ns, name):
    node_type, file = executable.split(":=")
    if "python" in node_type:
        if ".py" not in file:
            file = importlib.import_module(file).__file__
    p = subprocess.Popen([file] + [ns, name])
    return p


def initialize_nodes(
    nodes: Union[Union[Any, Dict], List[Union[Any, Dict]]],
    process_id: int,
    ns: str,
    message_broker: Any,
    is_initialized: Dict,
    sp_nodes: Dict,
    launch_nodes: Dict,
    rxnode_cls: Any = None,
    node_args: Dict = None,
    subscribers: List = None,
):
    if rxnode_cls is None:
        from eagerx.core.executable_node import RxNode

        rxnode_cls = RxNode

    from eagerx.core.specs import BaseNodeSpec

    if isinstance(nodes, (BaseNodeSpec, dict)):
        nodes = [nodes]

    for node in nodes:
        # Check if we still need to upload params to rosparam server (env)
        if not isinstance(node, dict):
            params = node.build(ns=ns)

            # Check if node name is unique
            name = node.config.name
            assert (
                rospy.get_param(("%s/%s/rate") % (ns, name), None) is None
            ), 'Node name "%s" already exists. Node names must be unique.' % (ns + "/" + name)

            # Upload params to rosparam server
            rosparam.upload_params(ns, params)

            # Make params consistent when directly grabbing params from rosparam server
            params = params[name]
        else:
            params = node
            name = params["name"]
        # Flag to check if node is initialized
        is_initialized[name] = False

        # Verify that node name is unique
        node_address = ns + "/" + name
        assert (node_address not in sp_nodes) and (node_address not in launch_nodes), (
            'Node "%s" already exists. Node names must be unique!' % name
        )

        # Block env until all nodes are initialized
        def initialized(msg, name):
            is_initialized[name] = True

        sub = rospy.Subscriber(node_address + "/initialized", UInt64, partial(initialized, name=name))
        message_broker.subscribers.append(sub)

        # Initialize node
        if params["process"] == process_id:  # Initialize inside this process
            if node_args is None:
                node_args = dict()
            sp_nodes[node_address] = rxnode_cls(name=node_address, message_broker=message_broker, **node_args)
            sp_nodes[node_address].node_initialized()
        elif (
            params["process"] == process.NEW_PROCESS and process_id == process.ENVIRONMENT
        ):  # Only environment can launch new processes (as it is the main_thread)
            assert "executable" in params, (
                'No executable defined. Node "%s" can only be launched as a separate process if an executable is specified.'
                % name
            )
            launch_nodes[node_address] = launch_node_as_subprocess(params["executable"], ns, name)
        elif params["process"] == process.EXTERNAL:
            rospy.loginfo('Node "%s" must be manually launched as the process is specified as process.EXTERNAL' % name)
        # else: node is launched in another (already launched) node's process (e.g. bridge process).


def wait_for_node_initialization(is_initialized, wait_time=0.3):
    iter = 0

    # Wait for nodes to be initialized
    while True:
        if iter == 0:
            sleep(0.1)
        else:
            sleep(wait_time)
        iter += 1
        not_init = []
        for name, flag in is_initialized.items():
            if not flag:
                not_init.append(name)
        if len(not_init) > 0:
            rospy.loginfo_once('Waiting for nodes "%s" to be initialized.' % (str(not_init)))
        else:
            break



# OTHER
from typing import Optional, Dict, Union, List
import gym

# ROS IMPORTS
import rospy
from std_msgs.msg import UInt64
from genpy.message import Message

# RX IMPORTS
from eagerx.core.constants import process, ERROR
import eagerx.core.register as register
from eagerx.core.entities import Bridge
from eagerx.core.specs import BridgeSpec


class GymBridge(Bridge):
    def initialize(self):
        self.simulator = dict()

    @staticmethod
    @register.spec("GymBridge", Bridge)
    def spec(
        spec: BridgeSpec,
        rate,
        process: Optional[int] = process.NEW_PROCESS,
        is_reactive: Optional[bool] = True,
        real_time_factor: Optional[float] = 0,
        simulate_delays: Optional[bool] = True,
        log_level: Optional[int] = ERROR,
    ):
        """TestBridge spec"""
        # Performs all the steps to fill-in the params with registered info about all functions.
        spec.initialize(GymBridge)

        # Modify default bridge params
        spec.config.rate = rate
        spec.config.process = process
        spec.config.is_reactive = is_reactive
        spec.config.real_time_factor = real_time_factor
        spec.config.simulate_delays = simulate_delays
        spec.config.log_level = log_level
        spec.config.color = "magenta"

    @register.bridge_config(env_id=None)
    def add_object(self, config, bridge_config, node_params, state_params):
        # add object to simulator (we have a ref to the simulator with self.simulator)
        rospy.loginfo(f'Adding object "{config["name"]}" of type "{config["entity_id"]}" to the simulator.')

        # Extract relevant object_params
        obj_name = config["name"]
        id = bridge_config["env_id"]

        # Create new env, and add to simulator
        self.simulator[obj_name] = dict(
            env=gym.make(id),
            buffer_obs=[],
            buffer_reward=None,
            buffer_done=None,
            next_action=None,
        )

    def pre_reset(self):
        pass

    @register.states()
    def reset(self):
        for _obj_name, sim in self.simulator.items():
            obs = sim["env"].reset()
            sim["buffer_obs"] = [obs]
            sim["buffer_reward"] = []
            sim["buffer_done"] = []

    @register.outputs(tick=UInt64)
    def callback(self, t_n: float, **kwargs: Dict[str, Union[List[Message], float, int]]):
        for _obj_name, sim in self.simulator.items():
            next_action = sim["next_action"]
            obs, reward, is_done, _ = sim["env"].step(next_action)
            sim["buffer_obs"].append(obs)
            sim["buffer_reward"].append(reward)
            sim["buffer_done"].append(is_done)



from typing import Optional
from threading import Event
import signal
import sys

import numpy as np
import rospy
from std_msgs.msg import UInt64, Bool
from sensor_msgs.msg import Image
from cv_bridge import CvBridge, CvBridgeError
import cv2


import eagerx.core.register as register
from eagerx.core.constants import process, WARN
from eagerx.core.entities import Node
from eagerx.core.specs import NodeSpec
from eagerx.utils.utils import initialize_converter, Msg


class EnvNode(Node):
    @staticmethod
    @register.spec("Environment", Node)
    def spec(spec: NodeSpec, rate=1, log_level=WARN, color="yellow"):
        """EnvNode Spec"""
        spec.initialize(EnvNode)

        # Modify default node params
        spec.config.name = "environment"
        spec.config.rate = rate
        spec.config.process = process.ENVIRONMENT
        spec.config.color = color
        spec.config.log_level = log_level
        spec.config.inputs = []
        spec.config.outputs = []
        spec.config.states = []

    def initialize(self):
        # Define observation buffers
        self.observation_buffer = dict()
        for i in self.inputs:
            if i["name"] == "actions_set":
                continue
            if "converter" in i and isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
                converter = i["converter"]
            elif "converter" in i and not isinstance(i["converter"], dict):
                converter = i["converter"]
            else:
                raise ValueError(f'Converter type {i["converter"]} of {i["name"]} not supported.')

            name = i["name"]
            window = i["window"]
            self.observation_buffer[name] = {
                "msgs": None,
                "converter": converter,
                "window": window,
            }

        # Synchronization event
        self.obs_event = Event()
        self.action_event = Event()
        self.must_reset = False

        # Graceful signal handler
        def signal_handler(sig, frame):
            print("SIGINT caught!")
            self.obs_event.set()
            self.action_event.set()
            sys.exit(0)

        signal.signal(signal.SIGINT, signal_handler)

        # Define action buffers
        self.action_buffer = dict()
        for i in self.outputs:
            if i["name"] == "set":
                continue
            if "converter" in i and isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
                converter = i["converter"]
            elif "converter" in i and not isinstance(i["converter"], dict):
                converter = i["converter"]
            else:
                converter = None
            self.action_buffer[i["name"]] = {"msg": None, "converter": converter}

    def reset(self):
        self.must_reset = False

        # Set all observation messages to None
        for _name, buffer in self.observation_buffer.items():
            buffer["msgs"] = None

        # Set all action messages to None
        for _name, buffer in self.action_buffer.items():
            buffer["msg"] = None

    def callback(self, t_n: float, **kwargs: Optional[Msg]):
        for name, i in kwargs.items():
            buffer = self.observation_buffer[name]
            window = buffer["window"]

            extra = window - len(i.msgs)
            if extra > 0:
                msgs = extra * [i.msgs[0]] + i.msgs
            else:
                msgs = i.msgs
            buffer["msgs"] = np.array(msgs)

        if not self.must_reset:
            self.action_event.clear()  # Clear action event, so that we can block after setting obs
            self.obs_event.set()  # Signal environment that observations have been set.
            try:
                # flag = self.action_event.wait(5)  # Wait for actions to be set.
                flag = self.action_event.wait()  # Wait for actions to be set.
                if not flag:
                    raise KeyboardInterrupt
            except (KeyboardInterrupt, SystemExit):
                print("[env] KEYBOARD INTERRUPT")
                raise

        if self.must_reset:
            return None
        else:
            # Fill output_msg with buffered actions
            output_msgs = dict()
            for name, buffer in self.action_buffer.items():
                output_msgs[name] = buffer["msg"]
            return output_msgs

    def shutdown(self):
        self.obs_event.set()


class ObservationsNode(Node):
    @staticmethod
    @register.spec("Observations", Node)
    def spec(spec: NodeSpec, rate=1, log_level=WARN, color="yellow"):
        """ObservationsNode spec"""
        # Initialize spec
        spec.initialize(ObservationsNode)

        # Modify default node params
        spec.config.name = "env/observations"
        spec.config.rate = rate
        spec.config.process = process.ENVIRONMENT
        spec.config.color = color
        spec.config.log_level = log_level
        spec.config.inputs = ["actions_set"]
        spec.config.outputs = ["set"]
        spec.config.states = []

        # Pre-set address to mock an actual input
        spec.inputs.actions_set.address = "env/actions/outputs/set"

    def initialize(self):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")

    def reset(self):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")

    @register.inputs(actions_set=UInt64)
    @register.outputs(set=UInt64)
    def callback(self, t_n: float, **kwargs: Optional[Msg]):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")


class ActionsNode(Node):
    @staticmethod
    @register.spec("Actions", Node)
    def spec(spec: NodeSpec, rate=1, log_level=WARN, color="yellow"):
        """ActionsNode spec"""
        # Initialize spec
        spec.initialize(ActionsNode)

        # Modify default node params
        spec.config.name = "env/actions"
        spec.config.rate = rate
        spec.config.process = process.ENVIRONMENT
        spec.config.color = color
        spec.config.log_level = log_level
        spec.config.inputs = ["step"]
        spec.config.outputs = ["set"]
        spec.config.states = []

        # Pre-set address to mock an actual input
        spec.inputs.step.address = "env/supervisor/outputs/step"

    def initialize(self):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")

    def reset(self):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")

    @register.inputs(observations_set=UInt64, step=UInt64)
    @register.outputs(set=UInt64)
    def callback(self, t_n: float, **kwargs: Optional[Msg]):
        raise NotImplementedError("This is a dummy class. Functionality is actually implemented in the Environment node.")


class RenderNode(Node):
    @staticmethod
    @register.spec("Render", Node)
    def spec(spec: NodeSpec, rate, display=True, log_level=WARN, color="grey"):
        """RenderNode spec"""
        # Initialize spec
        spec.initialize(RenderNode)

        # Modify default node params
        spec.config.name = "env/render"
        spec.config.rate = rate
        spec.config.process = process.ENVIRONMENT
        spec.config.color = color
        spec.config.log_level = log_level
        spec.config.inputs = ["image"]
        spec.config.outputs = ["done"]
        spec.config.states = []

        # Modify custom params
        spec.config.display = display

        # Pre-set window
        spec.inputs.image.window = 0

    def initialize(self, display):
        self.cv_bridge = CvBridge()
        self.window = None
        self.display = display
        self.last_image = Image(data=[])
        self.render_toggle = False
        self.window_closed = True
        rospy.Subscriber("%s/%s/toggle" % (self.ns, self.name), Bool, self._set_render_toggle)
        rospy.Subscriber("%s/%s/get_last_image" % (self.ns, self.name), Bool, self._get_last_image)
        self.pub_set_last_image = rospy.Publisher(
            "%s/%s/set_last_image" % (self.ns, self.name),
            Image,
            queue_size=0,
            latch=True,
        )

    def _set_render_toggle(self, msg):
        if msg.data:

            rospy.loginfo("START RENDERING!")
        else:
            rospy.loginfo("STOP RENDERING!")
        self.render_toggle = msg.data

    def _get_last_image(self, msg):
        self.pub_set_last_image.publish(self.last_image)

    def reset(self):
        pass
        # self.last_image = Image()

    @register.inputs(image=Image)
    @register.outputs(done=UInt64)
    def callback(self, t_n: float, image: Optional[Msg] = None):
        if len(image.msgs) > 0:
            self.last_image = image.msgs[-1]
        empty = self.last_image.height == 0 or self.last_image.width == 0
        if not empty and self.display and self.render_toggle:
            try:
                try:
                    cv_image = self.cv_bridge.imgmsg_to_cv2(self.last_image, "bgr8")
                except ImportError as e:
                    rospy.logwarn_once("[%s] %s. Using numpy instead." % (self.ns_name, e))

                    if isinstance(self.last_image.data, bytes):
                        cv_image = np.frombuffer(self.last_image.data, dtype=np.uint8).reshape(
                            self.last_image.height, self.last_image.width, -1
                        )
                    else:
                        cv_image = np.array(self.last_image.data, dtype=np.uint8).reshape(
                            self.last_image.height, self.last_image.width, -1
                        )
                    if "rgb" in self.last_image.encoding:
                        cv_image = cv2.cvtColor(cv_image, cv2.COLOR_RGB2BGR)
            except CvBridgeError as e:
                rospy.logwarn(e)
                return dict(done=UInt64())
            cv2.imshow("Render", cv_image)
            cv2.waitKey(1)
            self.window_closed = False
        elif not self.window_closed:
            cv2.destroyWindow("Render")
            self.window_closed = True

        # Fill output_msg with 'done' output --> signals that we are done rendering
        output_msgs = dict(done=UInt64())
        return output_msgs

    def shutdown(self):
        rospy.logdebug(f"[{self.name}] {self.name}.shutdown() called.")
        cv2.destroyAllWindows()



import eagerx.bridges.test.bridge  # noqa: F401
import eagerx.bridges.test.converters  # noqa: F401
import eagerx.bridges.test.nodes  # noqa: F401
import eagerx.bridges.test.objects  # noqa: F401
import eagerx.bridges.test.enginestates  # noqa: F401



from eagerx.core.specs import BridgeSpec
from eagerx.core.graph import Graph
from eagerx.core.env import EagerEnv
from typing import Dict, Tuple, Callable
import numpy as np
import gym


def step_fn(prev_obs, obs, action, steps):
    info = dict()
    return obs, obs.pop("reward", 0.0)[0], obs.pop("done", False)[0], info


class EagerGym(EagerEnv):
    def __init__(
        self,
        name: str,
        rate: float,
        graph: Graph,
        bridge: BridgeSpec,
        step_fn: Callable = step_fn,
    ) -> None:
        super().__init__(name=name, rate=rate, graph=graph, bridge=bridge, step_fn=step_fn)
        # Flatten action spaces
        self._reduced_action_space = super(EagerGym, self).action_space
        self._flattened_action_space, self._actions_all_discrete = get_flattened_space(self._reduced_action_space)

        # Flatten & reduce observation spaces (remove 'reward' & 'done')
        obs_space = super(EagerGym, self).observation_space.__dict__["spaces"]
        obs_space.pop("reward", None)
        obs_space.pop("done", None)
        self._reduced_obs_space = gym.spaces.Dict(obs_space)
        self._flattened_obs_space, self._obs_all_discrete = get_flattened_space(self._reduced_obs_space)
        assert not self._obs_all_discrete, "Only continuous observations are currently supported."

    @property
    def observation_space(self):
        return self._flattened_obs_space

    @property
    def action_space(self):
        return self._flattened_action_space

    def unflatten_action(self, action):
        # Unflatten action
        if not isinstance(action, np.ndarray):  # Discrete space
            action = np.array([action])
        return gym.spaces.unflatten(self._reduced_action_space, action)

    def flatten_observation(self, obs):
        if not self._obs_all_discrete:
            return gym.spaces.flatten(self._reduced_obs_space, obs)
        else:
            raise ValueError("Only continuous observations are currently supported.")

    def step(self, action: np.ndarray) -> Tuple[np.ndarray, float, bool, Dict]:
        action = self.unflatten_action(action)

        # Apply action
        obs, reward, is_done, info = super(EagerGym, self).step(action)

        # Flatten observation
        obs = self.flatten_observation(obs)
        return obs, reward, is_done, info

    def reset(self):
        obs = super(EagerGym, self).reset()
        obs = self.flatten_observation(obs)
        return obs


def get_flattened_space(spaces):
    if isinstance(spaces, gym.spaces.dict.Dict):
        spaces = spaces.__dict__["spaces"]
    if not isinstance(spaces, dict):
        spaces = dict(spaces)
    # Check if all discrete or mixed (with continuous)
    all_discrete = True
    for _key, space in spaces.items():
        if isinstance(space, gym.spaces.Box) and not (space.dtype == "int64" and space.shape == (1,)):
            all_discrete = False
        elif isinstance(space, gym.spaces.MultiDiscrete):
            raise ValueError("MultiDiscrete space not supported.")
    # If all discrete & multiple discrete, initialize MultiDiscrete, else just discrete
    if all_discrete and len(spaces) > 1:
        multi = []
        for _key, space in spaces.items():
            multi.append(space.high[0] + 1)
        flattened_space = gym.spaces.MultiDiscrete(multi)
    elif all_discrete and len(spaces) == 1:
        key = list(spaces)[0]
        flattened_space = gym.spaces.Discrete(spaces[key].high[0] + 1)
    else:
        for key, space in spaces.items():
            if isinstance(space, gym.spaces.Box) and space.dtype == "int64" and space.shape == (1,):
                spaces[key] = gym.spaces.Box(
                    low=space.low.astype("float32"),
                    high=space.high.astype("float32") + 0.9999,
                    dtype="float32",
                )
        flattened_space = gym.spaces.flatten_space(gym.spaces.Dict(spaces))
    return flattened_space, all_discrete



from yaml import dump
import copy
from eagerx.utils.utils import is_supported_type


supported_types = (str, int, list, float, bool, dict)


def keys_exists(element, *keys):
    """
    Check if *keys (nested) exists in `element` (dict).
    """
    if not isinstance(element, dict):
        raise AttributeError("keys_exists() expects dict as first argument.")
    if len(keys) == 0:
        return True
        # raise AttributeError("keys_exists() expects at least two arguments, one given.")

    _element = element
    for key in keys:
        try:
            _element = _element[key]
        except KeyError:
            return False
    return True


def get_dict(params, depth):
    for k in depth:
        params = params[k]
    return params


def _is_supported_type(param):
    try:
        is_supported_type(param, supported_types, none_support=True)
    except TypeError:
        if isinstance(param, (SpecView, GraphView)):
            param = param.to_dict()
            is_supported_type(param, supported_types, none_support=True)
        else:
            from eagerx.core.specs import EntitySpec

            if isinstance(param, EntitySpec):
                param = param.params
                is_supported_type(param, supported_types, none_support=True)
            else:
                raise


def _convert_type(param):
    if type(param) in supported_types or param is None:
        return param
    else:
        if isinstance(param, (SpecView, GraphView)):
            return param.to_dict()
        else:
            from eagerx.core.specs import EntitySpec

            if isinstance(param, EntitySpec):
                return param.params
            else:
                message = f'Type "{type(param)}" is not supported. Only types {supported_types} are supported.'
                raise TypeError(message)


class ViewIterator(object):
    def __init__(self, view, items=False):
        self._view = view
        self._items = items
        self._view_keys = list(view.keys()).copy()
        self._index = 0

    def __iter__(self):
        """Returns the Iterator object"""
        return self

    def __next__(self):
        if not len(self._view_keys) == len(self._view.keys()):
            raise ValueError("Iterator length not constant while iterating.")
        if self._index < len(self._view_keys):
            key = self._view_keys[self._index]
            self._index += 1
            if self._items:
                return key, self._view[key]
            else:
                return key
        raise StopIteration


class View(object):
    def __init__(self):
        super(View, self).__setattr__("_unlocked", False)

    def __getitem__(self, name):
        try:
            return self.__getattr__(name)
        except AttributeError:
            raise KeyError(name)

    def __setitem__(self, name, value):
        if name.startswith("__"):
            raise AttributeError("Cannot set magic attribute '{}'".format(name))
        self.__setattr__(name, value)

    def __iter__(self):
        return ViewIterator(self, items=False)

    def __enter__(self):
        self._unlock()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self._lock()

    @property
    def unlocked(self):
        """
        Use as with .unlocks as s: now set attributes to 's' with 's.<attribute>' within the context
        """
        return self

    def _lock(self):
        super(View, self).__setattr__("_unlocked", False)

    def _unlock(self):
        super(View, self).__setattr__("_unlocked", True)

    def update(self, mapping):
        for key, value in mapping.items():
            setattr(self, key, value)
        return self

    def items(self):
        return ViewIterator(self, items=True)


class SpecView(View):
    def __init__(self, spec, depth, name=None):
        super(SpecView, self).__setattr__("_spec", spec)
        super(SpecView, self).__setattr__("_depth", depth)
        super(SpecView, self).__setattr__("_name", name)
        super(SpecView, self).__init__()
        keys_exists(self._spec._params, *self._depth)

    def __setattr__(self, name, value):
        if self._spec.has_graph:
            raise AttributeError(f"Root Spec of '{self()}' was added to a graph, so this object has been disposed.")
        # Check if type is supported
        _is_supported_type(value)
        value = _convert_type(value)  # Convert Lookup & EntitySpec to dicts

        d = get_dict(self._spec._params, self._depth)
        if name in d:
            d[name] = value
        elif self._unlocked:
            d[name] = value
        else:
            message = f"Cannot set new attribute '{name}' when locked. \n"
            message += "The lock is there because new attributes are likely not accepted during initialization. "
            message += "If you are sure it will not break functionality, you can use '.unlock() before setting."
            raise AttributeError(message)

    def __getattr__(self, name):
        if self._spec.has_graph:
            raise AttributeError(f"Root Spec of '{self()}' was added to a graph, so this object has been disposed.")
        if keys_exists(self._spec._params, *self._depth, name):
            new_depth = self._depth.copy() + [name]
            d = get_dict(self._spec._params, new_depth)
            if isinstance(d, dict):
                return SpecView(self._spec, new_depth, self._name)
            else:
                return d
        else:
            try:
                d = get_dict(self._spec._params, self._depth)
                return getattr(d, name)
            except AttributeError as e:
                d = get_dict(self._spec._params, self._depth)
                if self._name:
                    depth_str = f"{self._name}"
                else:
                    depth_str = "params"
                for i in self._depth:
                    depth_str += f".{i}"
                attr_available = ["." + n for n in list(d.keys())]
                message = f"Attribute '{name}' not found. Available attributes in {depth_str}={attr_available}."
                raise AttributeError(message) from e

    def __str__(self):
        if self._spec.has_graph:
            raise AttributeError(f"Root Spec of '{self()}' was added to a graph, so this object has been disposed.")
        d = get_dict(self._spec._params, self._depth)
        repr = f"Params for {self()}: \n\n"
        repr += dump(d)
        return repr

    def __repr__(self):
        if self._spec.has_graph:
            raise AttributeError(f"Root Spec of '{self()}' was added to a graph, so this object has been disposed.")
        d = get_dict(self._spec._params, self._depth)
        repr = f"Params for {self()}: \n\n"
        repr += dump(d)
        return repr

    def __len__(self):
        return len(get_dict(self._spec._params, self._depth))

    def __call__(self):
        if self._name:
            return tuple([self._name] + self._depth)
        else:
            return tuple(self._depth)

    def to_dict(self):
        if self._spec.has_graph:
            raise AttributeError(f"Root Spec of '{self()}' was added to a graph, so this object has been disposed.")
        return get_dict(self._spec._params, self._depth)


class GraphView(View):
    def __init__(self, graph, depth, name):
        super(GraphView, self).__setattr__("_graph", graph)
        super(GraphView, self).__setattr__("_depth", depth)
        super(GraphView, self).__setattr__("_name", name)
        super(GraphView, self).__init__()
        keys_exists(self._graph._state["nodes"], *self._depth)

    def __setattr__(self, name, value):
        raise AttributeError(f"GraphView for '{self()}' is locked. Use graph API for setting once added to the graph.")

    def __getattr__(self, name):
        if keys_exists(self._graph._state["nodes"], *self._depth, name):
            new_depth = self._depth.copy() + [name]
            d = get_dict(self._graph._state["nodes"], new_depth)
            if isinstance(d, dict):
                return GraphView(self._graph, new_depth, self._name)
            else:
                return copy.deepcopy(d)
        else:
            try:
                d = get_dict(self._graph._state["nodes"], self._depth)
                return getattr(d, name)
            except AttributeError as e:
                d = get_dict(self._graph._state["nodes"], self._depth)
                if self._name:
                    depth_str = f"Spec({self._name})"
                else:
                    depth_str = "params"
                for i in self._depth[1:]:
                    depth_str += f".{i}"
                attr_available = ["." + n for n in list(d.keys())]
                message = f"Attribute '{name}' not found. Available attributes in {depth_str}={attr_available}."
                raise AttributeError(message) from e

    def __str__(self):
        d = get_dict(self._graph._state["nodes"], self._depth)
        repr = f"Params for {self()}: \n\n"
        repr += dump(d)
        return repr

    def __repr__(self):
        d = get_dict(self._graph._state["nodes"], self._depth)
        repr = f"Params for {self()}: \n\n"
        repr += dump(d)
        return repr

    def __call__(self):
        return tuple(self._depth)

    def __len__(self):
        return len(get_dict(self._graph._state["nodes"], self._depth))

    def to_dict(self):
        return copy.deepcopy(get_dict(self._graph._state["nodes"], self._depth))



from typing import Union, List
from std_msgs.msg import Float32MultiArray

from eagerx.core import register as register
from eagerx.core.entities import Processor


class GetIndex_Float32MultiArray(Processor):
    MSG_TYPE = Float32MultiArray

    @staticmethod
    @register.spec("GetIndex_Float32MultiArray", Processor)
    def spec(spec, index: Union[int, List[int]]):
        # Initialize spec with default arguments
        spec.initialize(GetIndex_Float32MultiArray)

        if isinstance(index, int):
            index = [index]
        for i in index:
            assert isinstance(i, int) and i >= 0, f'Incorrect index "{i}". Indices must be >= 0.'
        spec.config.index = index

    def initialize(self, index):
        self.index = index

    def convert(self, msg):
        if msg.data == []:
            return msg
        data = msg.data
        if isinstance(data, list) and len(data) > self.index:
            data = [data[self.index]]
        return Float32MultiArray(data=data)



from __future__ import print_function

# ROS imports
import rospy
import rosparam
from std_msgs.msg import UInt64, String, Bool
from sensor_msgs.msg import Image

# Rx imports
from eagerx.core.constants import process
from eagerx.core.executable_node import RxNode
import eagerx.core.rx_message_broker
import eagerx.core.rx_operators
import eagerx.core.rx_pipelines
from eagerx.core.entities import BaseNode
from eagerx.core.specs import NodeSpec, ObjectSpec
from eagerx.utils.utils import (
    get_attribute_from_module,
    initialize_converter,
    get_param_with_blocking,
)
from eagerx.utils.node_utils import initialize_nodes
from eagerx.core.nodes import EnvNode
import eagerx

# OTHER
from threading import Event


class SupervisorNode(BaseNode):
    def __init__(self, ns, states, **kwargs):
        self.subjects = None
        self.env_node: EnvNode = None

        # Render
        self.last_image = None
        self._image_event = Event()
        self.render_toggle = False
        self.pub_get_last_image = rospy.Publisher("%s/env/render/get_last_image" % ns, Bool, queue_size=0, latch=True)
        self.sub_set_last_image = rospy.Subscriber("%s/env/render/set_last_image" % ns, Image, self._last_image_callback)
        self.render_toggle_pub = rospy.Publisher("%s/env/render/toggle" % ns, Bool, queue_size=0, latch=True)

        # Initialize nodes
        self.cum_registered = 0
        self.is_initialized = dict()
        self.launch_nodes = dict()
        self.sp_nodes = dict()

        # Initialize buffer to hold desired reset states
        self.state_buffer = dict()
        for i in states:
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
                converter = i["converter"]
            else:
                converter = i["converter"]
            self.state_buffer[i["name"]] = {"msg": None, "converter": converter}

        # Required for reset
        self._step_counter = 0
        super().__init__(ns=ns, states=states, **kwargs)

    def set_environment(self, env_node: EnvNode):
        self.env_node = env_node

    def _set_subjects(self, subjects):
        self.subjects = subjects

    def start_render(self):
        if not self.render_toggle:
            self.render_toggle = True
            self.render_toggle_pub.publish(Bool(data=self.render_toggle))

    def stop_render(self):
        if self.render_toggle:
            self.render_toggle = False
            self.render_toggle_pub.publish(Bool(data=self.render_toggle))

    def get_last_image(self):
        self._image_event.clear()
        self.pub_get_last_image.publish(Bool())
        self._image_event.wait()
        return self.last_image

    def _last_image_callback(self, msg):
        self.last_image = msg
        self._image_event.set()

    def register_node(self, node: NodeSpec):
        # Increase cumulative registered counter. Is send as '/start_reset' message.
        self.cum_registered += 1

        # Initialize node
        node_name = node.config.name
        initialize_nodes(
            node,
            process.ENVIRONMENT,
            self.ns,
            self.message_broker,
            self.is_initialized,
            self.sp_nodes,
            self.launch_nodes,
            rxnode_cls=RxNode,
        )
        self.subjects["register_node"].on_next(String(self.ns + "/" + node_name))

    def register_object(self, object: ObjectSpec, bridge_name: str):
        # Increase cumulative registered counter. Is send as '/start_reset' message.
        self.cum_registered += 1

        # Check if object name is unique
        obj_name = object.config.name
        assert (
            rospy.get_param(self.ns + "/" + obj_name + "/nodes", None) is None
        ), f'Object name "{self.ns}/{obj_name}" already exists. Object names must be unique.'

        # Upload object params to rosparam server
        params, nodes = object.build(ns=self.ns, bridge_id=bridge_name)
        rosparam.upload_params(self.ns, params)

        # Upload node parameters to ROS param server
        initialize_nodes(
            nodes,
            process.ENVIRONMENT,
            self.ns,
            message_broker=self.message_broker,
            is_initialized=self.is_initialized,
            sp_nodes=self.sp_nodes,
            launch_nodes=self.launch_nodes,
        )
        self.subjects["register_object"].on_next(String(self.ns + "/" + obj_name))

    def _get_states(self, reset_msg):
        # Fill output_msg with buffered states
        msgs = dict()
        for name, buffer in self.state_buffer.items():
            if buffer["msg"] is None:
                msgs[name + "/done"] = Bool(data=True)
            else:
                msgs[name + "/done"] = Bool(data=False)
                msgs[name] = buffer["msg"]
                buffer["msg"] = None  # After sending state, set msg to None
        return msgs

    def reset(self):
        self.env_node.obs_event.clear()
        self.env_node.must_reset = True
        self.env_node.action_event.set()
        self.subjects["start_reset"].on_next(UInt64(data=self.cum_registered))
        self._step_counter = 0
        try:
            flag = self.env_node.obs_event.wait()
            if not flag:
                raise KeyboardInterrupt
        except (KeyboardInterrupt, SystemExit):
            print("[reset] KEYBOARD INTERRUPT")
            raise
        rospy.logdebug("FIRST OBS RECEIVED!")

    def step(self):
        self.env_node.obs_event.clear()
        self.env_node.action_event.set()
        self._step_counter += 1
        try:
            flag = self.env_node.obs_event.wait()
            if not flag:
                raise KeyboardInterrupt
        except (KeyboardInterrupt, SystemExit):
            print("[step] KEYBOARD INTERRUPT")
            raise
        rospy.logdebug("STEP END")

    def shutdown(self):
        self.env_node.action_event.set()
        self.pub_get_last_image.unregister()
        self.sub_set_last_image.unregister()
        self.render_toggle_pub.unregister()


class Supervisor(object):
    def __init__(self, name, message_broker, is_reactive, real_time_factor, simulate_delays):
        self.name = name
        self.ns = "/".join(name.split("/")[:2])
        self.mb = message_broker
        self.initialized = False
        self.is_reactive = is_reactive
        self.has_shutdown = False

        # Prepare input & output topics
        outputs, states, self.node = self._prepare_io_topics(self.name, is_reactive, real_time_factor, simulate_delays)

        # Initialize reactive pipeline
        rx_objects, env_subjects = eagerx.core.rx_pipelines.init_supervisor(
            self.ns, self.node, outputs=outputs, state_outputs=states
        )
        self.node._set_subjects(env_subjects)
        self.mb.add_rx_objects(node_name=name, node=self, **rx_objects)

    def node_initialized(self):
        # Notify env that node is initialized
        self.init_pub = rospy.Publisher(self.name + "/initialized", UInt64, queue_size=0, latch=True)
        self.init_pub.publish(UInt64())

        if not self.initialized:
            rospy.loginfo('Node "%s" initialized.' % self.name)
        self.initialized = True

    def _prepare_io_topics(self, name, is_reactive, real_time_factor, simulate_delays):
        params = get_param_with_blocking(name)

        # Get node
        node_cls = get_attribute_from_module(params["node_type"])
        node = node_cls(
            ns=self.ns,
            message_broker=self.mb,
            is_reactive=is_reactive,
            real_time_factor=real_time_factor,
            simulate_delays=simulate_delays,
            **params,
        )

        # Prepare output topics
        for i in params["outputs"]:
            i["msg_type"] = get_attribute_from_module(i["msg_type"])
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])

        # Prepare state topics
        for i in params["states"]:
            i["msg_type"] = get_attribute_from_module(i["msg_type"])
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])

        return tuple(params["outputs"]), tuple(params["states"]), node

    def _shutdown(self):
        rospy.logdebug(f"[{self.name}] Supervisor._shutdown() called.")
        self.init_pub.unregister()

    def node_shutdown(self):
        if not self.has_shutdown:
            rospy.logdebug(f"[{self.name}] Supervisor.node_shutdown() called.")
            rospy.loginfo(f"[{self.name}] Shutting down.")
            self._shutdown()
            self.node.shutdown()
            self.mb.shutdown()
            self.has_shutdown = True



#!/usr/bin/env python3

# ROS imports
import rospy
from std_msgs.msg import UInt64

# Rx imports
from eagerx.core.constants import log_levels_ROS
import eagerx.core.rx_message_broker
import eagerx.core.rx_pipelines
from eagerx.utils.utils import (
    get_attribute_from_module,
    initialize_converter,
    get_param_with_blocking,
    get_opposite_msg_cls,
)

# Other imports
import sys


class RxNode(object):
    def __init__(self, name, message_broker, **kwargs):
        self.name = name
        self.ns = "/".join(name.split("/")[:2])
        self.mb = message_broker
        self.initialized = False
        self.has_shutdown = False

        # Prepare input & output topics
        (
            rate,
            inputs,
            outputs,
            feedthroughs,
            states,
            targets,
            self.node,
        ) = self._prepare_io_topics(self.name, **kwargs)

        # Initialize reactive pipeline
        rx_objects = eagerx.core.rx_pipelines.init_node(
            self.ns,
            rate,
            self.node,
            inputs,
            outputs,
            feedthrough=feedthroughs,
            state_inputs=states,
            targets=targets,
        )
        self.mb.add_rx_objects(node_name=name, node=self, **rx_objects)

        # Prepare closing routine
        rospy.on_shutdown(self.node_shutdown)

    def node_initialized(self):
        # Notify env that node is initialized
        self.init_pub = rospy.Publisher(self.name + "/initialized", UInt64, queue_size=0, latch=True)
        self.init_pub.publish(UInt64())

        if not self.initialized:
            rospy.loginfo('Node "%s" initialized.' % self.name)
        self.initialized = True

    def _prepare_io_topics(self, name, **kwargs):
        params = get_param_with_blocking(name)
        rate = params["rate"]

        # Get info from bridge on reactive properties
        is_reactive = get_param_with_blocking(self.ns + "/bridge/is_reactive")
        real_time_factor = get_param_with_blocking(self.ns + "/bridge/real_time_factor")
        simulate_delays = get_param_with_blocking(self.ns + "/bridge/simulate_delays")

        # Get node
        node_cls = get_attribute_from_module(params["node_type"])
        node = node_cls(
            ns=self.ns,
            message_broker=self.mb,
            is_reactive=is_reactive,
            real_time_factor=real_time_factor,
            simulate_delays=simulate_delays,
            **kwargs,
            **params,
        )

        # Prepare input topics
        for i in params["inputs"]:
            i["msg_type"] = get_attribute_from_module(i["msg_type"])
            i["msg_type"] = get_opposite_msg_cls(i["msg_type"], i["converter"])
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
            # else:  # Converter already initialized

        # Prepare output topics
        for i in params["outputs"]:
            i["msg_type"] = get_attribute_from_module(i["msg_type"])
            i["msg_type"] = get_opposite_msg_cls(i["msg_type"], i["converter"])
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
            # else:  # Converter already initialized

        # Prepare state topics
        for i in params["states"]:
            i["msg_type"] = get_attribute_from_module(i["msg_type"])
            i["msg_type"] = get_opposite_msg_cls(i["msg_type"], i["converter"])
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
            # else:  # Converter already initialized

        # Prepare target topics
        for i in params["targets"]:
            i["msg_type"] = get_attribute_from_module(i["msg_type"])
            i["msg_type"] = get_opposite_msg_cls(i["msg_type"], i["converter"])
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
            # else:  # Converter already initialized

        # Prepare feedthrough topics
        for i in params["feedthroughs"]:
            i["msg_type"] = get_attribute_from_module(i["msg_type"])
            i["msg_type"] = get_opposite_msg_cls(i["msg_type"], i["converter"])
            if isinstance(i["converter"], dict):
                i["converter"] = initialize_converter(i["converter"])
            # else:  # Converter already initialized

        return (
            rate,
            tuple(params["inputs"]),
            tuple(params["outputs"]),
            tuple(params["feedthroughs"]),
            tuple(params["states"]),
            tuple(params["targets"]),
            node,
        )

    def _shutdown(self):
        rospy.logdebug(f"[{self.name}] RxNode._shutdown() called.")
        self.init_pub.unregister()

    def node_shutdown(self):
        if not self.has_shutdown:
            rospy.logdebug(f"[{self.name}] RxNode.node_shutdown() called.")
            rospy.loginfo(f"[{self.name}] Shutting down.")
            self._shutdown()
            self.node.shutdown()
            self.mb.shutdown()
            self.has_shutdown = True


if __name__ == "__main__":
    try:
        executable, ns, name = sys.argv[0], sys.argv[-2], sys.argv[-1]

        log_level = get_param_with_blocking(ns + "/log_level")

        rospy.init_node(
            f"{name}".replace("/", "_"),
            log_level=log_levels_ROS[log_level],
            anonymous=True,
        )

        message_broker = eagerx.core.rx_message_broker.RxMessageBroker(owner=f"{ns}/{name}")

        pnode = RxNode(name=f"{ns}/{name}", message_broker=message_broker)

        message_broker.connect_io()

        pnode.node_initialized()

        rospy.spin()
    finally:
        if not pnode.has_shutdown:
            rospy.loginfo(f"[{ns}/{name}] Send termination signal to '{ns}/{name}'.")
            rospy.signal_shutdown(f"Terminating '{ns}/{name}'")



import rospy
import yaml
from copy import deepcopy
import matplotlib.pyplot as plt
import networkx as nx
from typing import List, Union, Dict, Tuple, Optional, Any
from eagerx.utils.utils import (
    supported_types,
    get_opposite_msg_cls,
    get_cls_from_string,
    get_opposite_msg_cls_v2,
    substitute_args,
    msg_type_error,
)
from eagerx.utils.network_utils import (
    episode_graph,
    plot_graph,
    color_nodes,
    color_edges,
    is_stale,
)
from eagerx.core.graph import merge
from eagerx.core.specs import NodeSpec, ConverterSpec, EntitySpec
from eagerx.core.view import GraphView

yaml.Dumper.ignore_aliases = lambda *args: True


class EngineGraph:
    def __init__(self, state: Dict):
        self._state = state

    def __str__(self):
        return yaml.dump(self._state)

    @classmethod
    def create(cls, actuators: Optional[List[Dict]] = None, sensors: Optional[List[Dict]] = None):
        nodes = []

        from eagerx.core.entities import EngineNode

        # Create actuator node
        spec = EngineNode.pre_make(None, None)
        spec.config.name = "actuators"
        nodes.append(spec)
        for cname, params in actuators.items():
            # Determine converted msg_type, based on user-defined input converter
            # If input_conv != id but EngineNode connection also requires a converter --> raise error.
            conv_msg_type = get_opposite_msg_cls_v2(params.msg_type, params.converter)
            spec.add_output(
                cname,
                msg_type=conv_msg_type,
                converter=ConverterSpec(params.converter.to_dict()),  # Set input converter as output converter
                space_converter=ConverterSpec(params.space_converter.to_dict()),
            )
            spec.add_input(
                cname,
                msg_type=params.msg_type,
                skip=params.skip,
                converter=ConverterSpec(params.converter.to_dict()),
                space_converter=ConverterSpec(params.space_converter.to_dict()),
            )
            spec.config.outputs.append(cname)

        # Create sensor node
        spec = EngineNode.pre_make(None, None)
        spec.config.name = "sensors"
        nodes.append(spec)
        for cname, params in sensors.items():
            conv_msg_type = get_opposite_msg_cls_v2(params.msg_type, params.converter)
            spec.config.inputs.append(cname)
            spec.add_input(
                cname,
                msg_type=conv_msg_type,
                converter=ConverterSpec(params.converter.to_dict()),
                space_converter=ConverterSpec(params.space_converter.to_dict()),
            )

        # Create a state
        state = dict(nodes=dict(), connects=list(), backup=dict())
        graph = cls(state)
        graph.add(nodes)
        return graph

    def add(self, nodes: Union[NodeSpec, List[NodeSpec]]):
        """
        Add a node/object to the provided state.
        """
        if not isinstance(nodes, list):
            nodes = [nodes]

        for node in nodes:
            name = node.config.name
            assert name not in self._state["nodes"], (
                'There is already a node or object registered in this graph with name "%s".' % name
            )
            assert not node.has_graph, f"NodeSpec '{name}' is already added to a graph."

            # Add node to state
            self._state["nodes"][name] = node.params
            self._state["backup"][name] = node.params

            # Add graph reference to spec
            node.set_graph(self)

    def remove(self, names: Union[Union[str, EntitySpec], List[Union[str, EntitySpec]]]):
        """
        Removes a node.
        First removes all associated connects from self._state.
        Then, removes node/object from self._state.
        """
        if not isinstance(names, list):
            names = [names]
        for name in names:
            if isinstance(name, EntitySpec):
                name = name.params["config"]["name"]
                assert name in self._state["nodes"], f" No entity with name '{name}' in graph."
            for source, target in deepcopy(self._state["connects"]):
                if name in [source[0], target[0]]:
                    if source[0] == "actuators":
                        actuator = source[2]
                        source = None
                    else:
                        actuator = None
                        source = self.get_view(source[0], source[1:])
                    if target[0] == "sensors":
                        sensor = target[2]
                        target = None
                    else:
                        sensor = None
                        target = self.get_view(target[0], target[1:])
                    self.disconnect(source, target, actuator, sensor)
            self._state["nodes"].pop(name)

    def add_component(self, entry: GraphView):
        """
        Adds a component entry to the selection list.
        """
        name, component, cname = entry()
        # Add cname to selection list if it is not already selected
        params = self._state["nodes"][name]
        assert cname not in params["config"][component], f'"{cname}" already selected in "{name}" under {component}.'
        params["config"][component].append(cname)

    def remove_component(self, entry: GraphView):
        """
        Removes a component entry from the selection list. It will first disconnect all connections in connect.
        """
        self._is_selected(self._state, entry)

        # Disconnect component entry
        self._disconnect_component(entry)

        # Remove cname from selection list
        name, component, cname = entry()
        params = self._state["nodes"][name]
        params["config"][component].remove(cname)

    def connect(
        self,
        source: Optional[GraphView] = None,
        target: Optional[GraphView] = None,
        actuator: str = None,
        sensor: str = None,
        address: str = None,
        converter: Optional[Dict] = None,
        window: Optional[int] = None,
        delay: Optional[float] = None,
        skip: Optional[bool] = None,
        external_rate: Optional[float] = None,
    ):
        flag = not address or (source is None and actuator is None and sensor is None)
        assert flag, f'You cannot provide an external address "{address}" together with a sensor, actuator, or source.'
        flag = not source or not actuator
        assert flag, (
            f'You cannot specify an actuator if you wish to connect actuator "{actuator}", '
            "as the actuator will act as the source."
        )
        flag = not target or not sensor
        assert flag, (
            f'You cannot specify a target if you wish to connect sensor "{sensor}", ' "as the sensor will act as the target."
        )
        assert not (actuator and sensor), "You cannot connect an actuator directly to a sensor."
        if address:
            curr_address = target.address
            assert curr_address is None, (
                f'Cannot connect target "{target}" to external address "{address}", '
                f'because it is already connected to an external address "{curr_address}". '
                "Disconnect target first."
            )
            self.set({"address": address}, target)
            assert external_rate is not None, (
                f'When providing an external address "{address}", ' "an external rate must also be provided."
            )
            assert external_rate > 0, f'Invalid external rate "{external_rate}". External rate must be > 0.'
            self.set({"external_rate": external_rate}, target)
            return
        else:
            assert external_rate is None, "An external rate may only be provided in combination with an address."

        if isinstance(converter, ConverterSpec):
            converter = converter.params
        if isinstance(converter, GraphView):
            converter = converter.to_dict()

        if actuator:  # source = actuator
            source = self.get_view("actuators", ["outputs", actuator])
            if converter:
                msg = (
                    f'Cannot specify an input converter for actuator "{actuator}", '
                    "because one has already been specified in the agnostic graph definition. "
                    "You can only have one input converter."
                )
                from eagerx.core.converters import Identity

                assert source.converter.to_dict() == Identity().get_yaml_definition(), msg
            assert window is None, (
                f'Cannot specify a window when connecting actuator "{actuator}". '
                f"You can only do that in the agnostic object definition."
            )
            assert delay is None, (
                f'Cannot specify a delay when connecting actuator "{actuator}". '
                f"You can only do that in the agnostic object definition."
            )
            assert skip is None, (
                f'Cannot specify a skip when connecting actuator "{actuator}". '
                f"You can only do that in the agnostic object definition."
            )
        elif sensor:  # target = sensor
            target = self.get_view("sensors", ["inputs", sensor])
            # Check for output converter clash.
            # I.e. if both agnostic sensor & enginenode output have output converter.
            from eagerx.core.converters import Identity

            id = Identity().get_yaml_definition()
            if source.converter.to_dict() != id:
                src_name, _, src_cname = source()
                msg = (
                    "Output converter clash! "
                    f"Output '{src_cname}' of EngineNode '{src_name}' you attempt to connect to sensor '{sensor}'"
                    f" both have an output converter defined, but only one output converter can be used."
                )
                assert target.converter.to_dict() == id, msg
            assert converter is None, (
                f'Cannot specify an input converter when connecting sensor "{sensor}". '
                "For sensors, you can only do that in the agnostic definition. "
            )
            assert window is None, (
                f'Cannot specify a window when connecting sensor "{sensor}".'
                " You can only do that in the agnostic object definition."
            )
            assert delay is None, (
                f'Cannot specify a delay when connecting sensor "{sensor}".'
                " You can only do that in the agnostic object definition."
            )
            assert skip is None, (
                f'Cannot specify a skip when connecting sensor "{sensor}".'
                " You can only do that in the agnostic object definition."
            )
        self._connect(source, target, converter, window, delay, skip)

    def _connect(
        self,
        source: Optional[GraphView] = None,
        target: Optional[GraphView] = None,
        converter: Optional[Dict] = None,
        window: Optional[int] = None,
        delay: Optional[float] = None,
        skip: Optional[bool] = None,
    ):
        """
        Method to connect a source to a target. For actuators/sensors, first a (new) disconnected entry must be created,
        after which an additional call to connect_actuator/sensor is required before calling this method.
        For more info, see self.connect.
        """
        # Perform checks on source
        self._is_selected(self._state, source)

        # Perform checks on target
        self._is_selected(self._state, target)

        # Make sure that target is not already connected.
        curr_address = target.address
        assert curr_address is None, (
            f'Cannot connect target "{target}" to source "{source}", '
            f'because it is already connected to an external address "{curr_address}". '
            "Disconnect target first."
        )
        target_name, target_comp, target_cname = target()
        for s, t in self._state["connects"]:
            t_name, t_comp, t_cname = t
            flag = not (target_name == t_name and target_comp == t_comp and target_cname == t_cname)
            assert flag, f'Target "{target}" is already connected to source "{s}"'

        # Add properties to target params
        if converter is not None:
            self._set_converter(target, converter)
            # self.set({"converter": converter}, target)
        if window is not None:
            self.set({"window": window}, target)
        if delay is not None:
            self.set({"delay": delay}, target)
        if skip is not None:
            self.set({"skip": skip}, target)

        # Add connection
        EngineGraph.check_msg_type(source, target, self._state)
        connect = [list(source()), list(target())]
        self._state["connects"].append(connect)

    def disconnect(
        self,
        source: Optional[GraphView] = None,
        target: Optional[GraphView] = None,
        actuator: str = None,
        sensor: str = None,
    ):
        """
        Disconnects a source from a target. The target is reset in self._state to its disconnected state.
        """
        assert not source or not actuator, (
            "You cannot specify a source if you wish to disconnect actuator ",
            f'"{actuator}", as the actuator will act as the source.',
        )
        assert not target or not sensor, (
            "You cannot specify a target if you wish to disconnect sensor "
            f'"{sensor}", as the sensor will act as the target.'
        )
        assert not (sensor and actuator), "You cannot disconnect an actuator from a sensor, as such a connection cannot exist."

        # Create source & target entries
        if actuator:
            source = self.get_view("actuators", ["outputs", actuator])
        if sensor:
            target = self.get_view("sensors", ["inputs", sensor])

        self._disconnect(source, target)

    def _disconnect(self, source: GraphView, target: GraphView):
        """
        Disconnects a source from a target. The target is reset in self._state to its disconnected state.
        """
        # Check if connection exists
        self._is_selected(self._state, target)

        # Check if connection exists
        if source is None:
            flag = target.address is not None
            assert flag, (
                f"Cannot disconnect. No source was provided, and the target={target} "
                "also does not have an address specified."
            )
            self.set(None, target, parameter="address")
            self.set(None, target, parameter="external_rate")
        else:
            self._is_selected(self._state, source)
            connect_exists = False
            idx_connect = None
            for idx, c in enumerate(self._state["connects"]):
                if list(source()) == c[0] and list(target()) == c[1]:
                    connect_exists = True
                    idx_connect = idx
                    break
            assert connect_exists, (
                f"The connection with source={source()} and target={target()} cannot be removed," " because it does not exist."
            )

            # Pop the connection from the state
            self._state["connects"].pop(idx_connect)

        # Reset target params to disconnected state (reset to go back to default yaml), i.e. reset window/delay/skip/converter.
        target_name, target_comp, target_cname = target()
        target_params = self._state["nodes"][target_name]
        target_params[target_comp][target_cname] = self._state["backup"][target_name][target_comp][target_cname]

    def _disconnect_component(self, entry: GraphView):
        """
        Disconnects all associated connects from self._state.
        """
        was_connected = False
        name, component, cname = entry()
        for source, target in deepcopy(self._state["connects"]):
            source = self.get_view(source[0], source[1:])
            target = self.get_view(target[0], target[1:])
            self._is_selected(self._state, source)
            self._is_selected(self._state, target)
            source_name, source_comp, source_cname = source()
            target_name, target_comp, target_cname = target()
            if name == source_name and component == source_comp and cname == source_cname:
                self.disconnect(source, target)
                was_connected = True
            elif name == target_name and component == target_comp and cname == target_cname:
                self.disconnect(source, target)
                was_connected = True
        return was_connected

    def set(self, mapping: Any, entry: Optional[GraphView], parameter: Optional[str] = None):
        """
        Sets parameters in self._state, based on the node/object name. If a component and cname are specified, the
        parameter will be set there. Else, the parameter is set under the "config" key.
        For objects, parameters are set under their agnostic definitions of the components (so not bridge specific).
        If a converter is added, we check if the msg_type changes with the new converter. If so, the component is
        disconnected. See _set_converter for more info.
        """
        assert not entry()[0] == "actuators", (
            "Cannot change the actuator parameters here, "
            "in a bridge specific implementation. That is only possible in the "
            "object's agnostic definition."
        )
        assert not entry()[0] == "sensors", (
            "Cannot change the sensor parameters here, "
            "in a bridge specific implementation. That is only possible in the "
            "object's agnostic definition."
        )

        if parameter is None:
            if isinstance(mapping, GraphView):
                mapping = mapping.to_dict()
            assert isinstance(mapping, dict), "Can only set mappings of type dict. Else, also set 'parameter=<param_name>'."
        else:
            mapping = {parameter: mapping}
        for parameter, value in mapping.items():
            if parameter:
                getattr(entry, parameter)  # Check if parameter exists
            if parameter == "converter":
                msg = (
                    "Skipping converter. Cannot change the converter with this method. "
                    "Add output converters before connecting, and input converters when making a connection."
                )
                rospy.logwarn_once(msg)
            else:
                t = entry()
                name = t[0]
                if t[1] == "config":
                    assert parameter not in [
                        "sensors",
                        "actuators",
                        "targets",
                        "states",
                        "inputs",
                        "outputs",
                    ], "You cannot modify component parameters with this function. Use _add/remove_component(..) instead."
                    assert parameter not in ["msg_type"], f"You cannot rename '{name}'."
                    assert parameter not in ["msg_type"], f"You cannot modify msg_type '{value}'."
                    p = self._state["nodes"][name]["config"]
                else:
                    name, component, cname = entry()
                    p = self._state["nodes"][name][component][cname]
                self._set(p, {parameter: value})

    def _set_converter(self, entry: GraphView, converter: Dict):
        """
        Replaces the converter specified for a node's/object's I/O.
        **DOES NOT** remove observation entries if they are disconnected.
        **DOES NOT** remove action entries if they are disconnect and the last connection.
        """
        if isinstance(converter, ConverterSpec):
            converter = converter.params
        elif isinstance(converter, GraphView):
            converter = converter.to_dict()

        _ = entry.converter  # Check if parameter exists

        # Check if converted msg_type of old converter is equal to the msg_type of newly specified converter
        msg_type = get_cls_from_string(entry.msg_type)
        converter_old = entry.converter.to_dict()
        msg_type_ros_old = get_opposite_msg_cls(msg_type, converter_old)
        msg_type_ros_new = get_opposite_msg_cls(msg_type, converter)
        if not msg_type_ros_new == msg_type_ros_old:
            was_disconnected = self._disconnect_component(entry)
        else:
            was_disconnected = False

        # If disconnected action/observation, we cannot add converter so raise error.
        name, component, cname = entry()
        params = self._state["nodes"][name]
        flag = not (was_disconnected and name in ["actuators", "sensors"])
        assert flag, (
            f'Cannot change the converter of actuator/sensor "{cname}", '
            f'as it changes the msg_type from "{msg_type_ros_old}" to "{msg_type_ros_new}"'
        )

        # Replace converter
        params[component][cname]["converter"] = converter

    def get(
        self,
        entry: Optional[Union[GraphView, EntitySpec]] = None,
        actuator: Optional[str] = None,
        sensor: Optional[str] = None,
        parameter: Optional[str] = None,
    ):
        if isinstance(entry, EntitySpec):
            name = entry.params["config"]["name"]
            assert name in self._state["nodes"], f" No entity with name '{name}' in graph."
            return self._state["nodes"][name]
        self._correct_signature(entry, actuator, sensor)
        if actuator:
            entry = self.get_view("actuators", ["outputs", actuator])
        if sensor:
            entry = self.get_view("sensors", ["inputs", sensor])
        if parameter:
            return getattr(entry, parameter)
        else:
            return entry

    def _node_depenencies(self, state):
        import networkx as nx

        state = deepcopy(state)
        G = self._generate_graph(state)
        G_rev = G.reverse(copy=True)

        # fig_env, ax_env = plt.subplots(nrows=1, ncols=1)
        # ax_env.set_title('Engine-specific graph')
        # _, _, _, pos = plot_graph(G, k=2, ax=ax_env)
        # plt.show()
        #
        # fig_env, ax_env = plt.subplots(nrows=1, ncols=1)
        # ax_env.set_title('Engine-specific graph (reversed)')
        # _, _, _, _ = plot_graph(G_rev, k=2, ax=ax_env, pos=pos)
        # plt.show()

        # Determine sensor dependencies
        dependencies = dict(sensors=dict(), actuators=dict())
        for cname in state["nodes"]["sensors"]["inputs"]:
            dependencies["sensors"][cname] = []
            target_name = f"sensors/{cname}"
            descendants = nx.descendants(G_rev, target_name)
            for source in descendants:
                node_name, source_cname = source.split("/")
                if node_name in ["actuators", "sensors"]:
                    continue
                dependencies["sensors"][cname].append(node_name)

        # Determine actuator dependency
        for cname in state["nodes"]["actuators"]["outputs"]:
            dependencies["actuators"][cname] = []
            source_name = f"actuators/{cname}"
            descendants = nx.descendants(G, source_name)
            for target in descendants:
                node_name, target_cname = target.split("/")
                if node_name in ["actuators", "sensors"]:
                    continue
                dependencies["actuators"][cname].append(node_name)
            # Also add dependencies of all targets
            for target in descendants:
                nx.descendants(G_rev, target)
                for source in descendants:
                    node_name, source_cname = source.split("/")
                    if node_name in ["actuators", "sensors"]:
                        continue
                    dependencies["actuators"][cname].append(node_name)
            dependencies["actuators"][cname] = list(set(dependencies["actuators"].pop(cname)))
        return dependencies

    def register(self):
        """Set the addresses in all incoming components.
        Validate the graph.
        Create params that can be uploaded to the ROS param server.
        """

        # Check if valid graph.
        assert self.is_valid(plot=False), "Graph not valid."

        # Find dependencies
        dependencies = self._node_depenencies(self._state)

        # Add addresses based on connections
        state = deepcopy(self._state)
        actuators = dict()
        sensors = dict()
        for source, target in state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target
            address = EngineGraph._get_address(source, target)
            if source_name == "actuators":
                dependency = [f"$(ns obj_name)/{d}" for d in dependencies["actuators"][source_cname]]
                if source_cname not in actuators:
                    actuators[source_cname] = []
                entry = {
                    "name": f"$(ns obj_name)/{target_name}",
                    "component": target_comp,
                    "cname": target_cname,
                    "dependency": dependency,
                }
                # actuators[source_cname] = entry
                actuators[source_cname].append(entry)
                continue  # we continue here, because the address for actuators is determined by an output from the agnostic graph.
            if target_name == "sensors":
                dependency = [f"$(ns obj_name)/{d}" for d in dependencies["sensors"][target_cname]]
                if target_cname not in sensors:
                    sensors[target_cname] = []
                entry = {
                    "name": f"$(ns obj_name)/{source_name}",
                    "component": source_comp,
                    "cname": source_cname,
                    "dependency": dependency,
                }
                # sensors[target_cname] = entry
                sensors[target_cname].append(entry)
                continue  # we continue here, because the address for actuators is determined by an output from the agnostic graph.
            state["nodes"][target_name][target_comp][target_cname]["address"] = address

        # Initialize param objects
        nodes = dict()
        from eagerx.core.specs import NodeSpec

        for name, params in state["nodes"].items():
            if "node_type" in params:
                if name == "actuators":
                    pass
                elif name == "sensors":
                    pass
                else:
                    # Put node name into object namespace
                    spec = NodeSpec(params)
                    name = f"$(ns obj_name)/{spec.config.name}"
                    spec.config.name = name
                    params = spec.params

                    # Substitute placeholder args of simnode
                    context = {"ns": {"node_name": name}, "config": params["config"]}
                    substitute_args(params, context, only=["config", "ns"])
                    nodes[name] = params

        assert len(actuators) > 0, "No actuators node defined in the graph."
        assert len(sensors) > 0, "No sensors node defined in the graph."
        return nodes, actuators, sensors

    def gui(self):
        raise NotImplementedError("Gui is not yet supported for engine graphs.")

    @staticmethod
    def _get_address(source: Tuple[str, str, str], target: Tuple[str, str, str]):
        """Determine the address."""
        source_name, source_comp, source_cname = source
        target_name, target_comp, target_cname = target
        if source_name == "actuators":
            assert (
                not target_name == "sensors"
            ), f'A direct connection between a sensor "{target_cname}" and actuator "{source_cname}" cannot exist.'
            address = f"$(ns obj_name)/{source_name}/{source_cname}"
        elif target_name == "sensors":
            assert (
                not source_name == "actuators"
            ), f'A direct connection between a sensor "{target_cname}" and actuator "{source_cname}" cannot exist.'
            address = f"$(ns obj_name)/{target_name}/{target_cname}"
        else:
            address = f"$(ns obj_name)/{source_name}/{source_comp}/{source_cname}"
        return address

    @staticmethod
    def _is_selected(state: Dict, entry: GraphView):
        """
        Check if provided entry was selected in params.
        """
        name, component, cname = entry()
        params = state["nodes"][name]
        component = "outputs" if component == "feedthroughs" else component
        assert cname in params["config"][component], f'"{cname}" not selected in "{name}" under "config" in {component}.'

    @staticmethod
    def _correct_signature(
        entry: Optional[GraphView] = None,
        actuator: Optional[str] = None,
        sensor: Optional[str] = None,
    ):
        # assert only action, only observation, or only entry
        if entry:  # component parameter
            assert actuator is None, "If 'entry' is specified, actuator argument cannot be specified."
            assert sensor is None, "If 'entry' is specified, sensor argument cannot be specified."
        if actuator:
            assert sensor is None, "If actuator is specified, sensor must be None."
            assert entry is None, "If actuator is specified, the 'entry' argument cannot be specified."
        if sensor:
            assert actuator is None, "If sensor is specified, action must be None."
            assert entry is None, "If actuator is specified, the 'entry' argument cannot be specified."

    def is_valid(self, plot=True):
        return self._is_valid(self._state, plot=plot)

    @staticmethod
    def _is_valid(state, plot=True):
        state = deepcopy(state)
        EngineGraph.check_msg_types_are_consistent(state)
        EngineGraph.check_inputs_have_address(state)
        EngineGraph.check_graph_is_acyclic(state, plot=plot)
        return True

    @staticmethod
    def check_msg_type(source, target, state):
        # Convert the source msg_type to target msg_type with converters:
        # msg_type_source --> output_converter --> msg_type_ROS --> input_converter --> msg_type_target
        msg_type_out = get_cls_from_string(source.msg_type)
        converter_out = source.converter.to_dict()
        msg_type_ros = get_opposite_msg_cls(msg_type_out, converter_out)
        converter_in = target.converter.to_dict()
        msg_type_in = get_opposite_msg_cls(msg_type_ros, converter_in)

        # Verify that this msg_type_in is the same as the msg_type specified in the target
        msg_type_in_registered = get_cls_from_string(target.msg_type)

        msg_type_str = msg_type_error(
            source(),
            target(),
            msg_type_out,
            converter_out,
            msg_type_ros,
            converter_in,
            msg_type_in,
            msg_type_in_registered,
        )
        assert msg_type_in == msg_type_in_registered, msg_type_str

    @staticmethod
    def check_msg_types_are_consistent(state):
        for source, target in state["connects"]:
            if target[0] == "sensors":
                target_view = EngineGraph._get_view(state, target[0], target[1:])
                from eagerx.core.converters import Identity

                id = Identity().get_yaml_definition()
                # If agnostic definition has a non-identity output_converter,
                # enginenode output cannot be inter-connected with other enginenodes,
                # as they expect an unconverted output.
                if target_view.converter.to_dict() != id:
                    flag = [t for s, t in state["connects"] if s == source]
                    msg = (
                        f"Agnostic definition for sensor '{target[2]}' has a non-identity output converter, "
                        f"However, output '{source[2]}' of EngineNode '{source[0]}' is interconnected to other "
                        f"EngineNodes that expect the original (unconverted) output message: {flag}. "
                        f"Non-identity output converters can only be added to a sensor if the corresponding "
                        f"EngineNode output has only a single connection (with the sensor)."
                    )
                    assert len(flag) == 1, msg
        for source, target in state["connects"]:
            source = EngineGraph._get_view(state, source[0], source[1:])
            target = EngineGraph._get_view(state, target[0], target[1:])
            EngineGraph.check_msg_type(source, target, state)
        return True

    @staticmethod
    def check_inputs_have_address(state):
        state = deepcopy(state)
        for source, target in state["connects"]:
            address = EngineGraph._get_address(source, target)
            target_name, target_comp, target_cname = target
            state["nodes"][target_name][target_comp][target_cname]["address"] = address

        for name, params in state["nodes"].items():
            for component in params["config"]:
                if component not in [
                    "inputs",
                    "outputs",
                    "targets",
                    "feedthroughs",
                    "states",
                ]:
                    continue
                for cname in params["config"][component]:
                    flag = cname in params[component]
                    assert flag, f'"{cname}" was selected in {component} of "{name}", but has no implementation.'
                    if component not in ["inputs", "targets", "feedthroughs"]:
                        continue
                    if name in ["sensors", "actuators"]:
                        continue
                    flag = params[component][cname]["address"] is not None
                    assert flag, (
                        f'"{cname}" was selected in {component} of "{name}", but no address was specified. '
                        f"Either deselect it, or connect it."
                    )
        return True

    @staticmethod
    def check_graph_is_acyclic(state, plot=True):
        # Generate graph
        G = EngineGraph._generate_graph(state)

        # Color nodes based on in/out going edges
        not_active = is_stale(G, exclude_skip=True)
        color_nodes(G)
        color_edges(G)

        # Check if graph is acyclic (excluding 'skip' edges)
        H, cycles = episode_graph(G)
        is_dag = nx.is_directed_acyclic_graph(H)

        # Plot graphs
        if plot:
            fig_env, ax_env = plt.subplots(nrows=1, ncols=1)
            ax_env.set_title("Engine-specific graph")
            _, _, _, pos = plot_graph(G, k=2, ax=ax_env)
            plt.show()

        # Assert if graph is a directed-acyclical graph (DAG)
        cycle_strs = ["Algebraic loops detected: "]
        for idx, connect in enumerate(cycles):
            connect.append(connect[0])
            s = " Loop %s: " % idx
            n = "\n" + "".join([" "] * len(s)) + "...-->"
            s = "\n\n" + s + "...-->"
            for idx in range(len(connect) - 1):
                tmp, target = connect[idx]
                source, tmp2 = connect[idx + 1]
                source_name, source_comp, source_cname = source
                target_name, target_comp, target_cname = target
                assert source_name == target_name, "Source and target not equal: %s, %s" % (source, target)
                connect_name = "%s/%s/%s][%s/%s/%s" % tuple(list(source) + list(target))
                node_name = ("Node: " + source_name).center(len(connect_name), " ")
                s += "[%s]-->" % connect_name
                n += "[%s]-->" % node_name
            s += "..."
            n += "..."
            cycle_strs.append(s)
            cycle_strs.append(n)
            connect.pop(-1)
        assert is_dag, "".join(cycle_strs)

        not_active_excl_sensors = [n for n in not_active if not n.split("/")[0] == "sensors"]
        assert len(not_active_excl_sensors) == 0, (
            'Stale episode graph detected. Nodes "%s" will be stale, while they must be active (i.e. connected) in order for the graph to resolve (i.e. not deadlock).'
            % not_active
        )
        return True

    @staticmethod
    def _generate_graph(state):
        # Add nodes
        G = nx.MultiDiGraph()
        # label_mapping = {'env/observations/set': 'observations', 'env/render/done': 'render'}
        for node, params in state["nodes"].items():
            default = params["config"]
            if "outputs" in default:
                has_tick = True if "tick" in default["inputs"] else False
                for cname in default["outputs"]:
                    name = "%s/%s" % (node, cname)
                    remain_active = True if node == "actuators" else False
                    always_active = True if node == "actuators" else False
                    G.add_node(
                        name,
                        remain_active=remain_active,
                        always_active=always_active,
                        is_stale=False,
                        has_tick=has_tick,
                    )
            if node == "sensors":
                for cname in default["inputs"]:
                    name = "%s/%s" % (node, cname)
                    G.add_node(
                        name,
                        remain_active=True,
                        always_active=False,
                        is_stale=False,
                        has_tick=False,
                    )

        # Add edges
        target_comps = ["inputs"]
        source_comps = ["outputs"]
        for source, target in state["connects"]:
            source_name, source_comp, source_cname = source
            target_name, target_comp, target_cname = target
            if source_comp in source_comps and target_comp in target_comps:
                source_edge = "%s/%s" % (source_name, source_cname)
                # Determine target node name
                target_edges = []
                target_default = state["nodes"][target_name]["config"]
                for cname in target_default["outputs"]:
                    target_edge = "%s/%s" % (target_name, cname)
                    target_edges.append(target_edge)
                if target_name == "sensors":
                    target_edge = "%s/%s" % (target_name, target_cname)
                    target_edges.append(target_edge)

                # Determine edges that do not break DAG property (i.e. edges that are skipped)
                if source_name == "actuators":
                    skip = state["nodes"][source_name]["inputs"][source_cname]["skip"]
                else:
                    skip = state["nodes"][target_name][target_comp][target_cname]["skip"]

                # Determine color
                color = "green" if skip else "black"
                style = "dotted" if skip else "solid"

                # Add edge
                for target_edge in target_edges:
                    G.add_edge(
                        source_edge,
                        target_edge,
                        color=color,
                        feedthrough=False,
                        style=style,
                        alpha=1.0,
                        is_stale=False,
                        skip=skip,
                        source=source,
                        target=target,
                    )

        # Color nodes based on in/out going edges
        # not_active = is_stale(G, exclude_skip=True)
        color_nodes(G)
        color_edges(G)
        return G

    @staticmethod
    def _get_view(state, name: str, depth: Optional[List[str]] = None):
        depth = depth if depth else []
        depth = [name] + depth
        return GraphView(EngineGraph(state), depth=depth, name=name)

    def get_view(self, name: str, depth: Optional[List[str]] = None):
        depth = depth if depth else []
        depth = [name] + depth
        return GraphView(self, depth=depth, name=name)

    @staticmethod
    @supported_types(str, int, list, float, bool, dict, EntitySpec, GraphView, None)
    def _set(state, mapping):
        merge(state, mapping)
